import copy
import datetime
import functools
import itertools
import json

# setup ES logging
import logging
import re
import time
from importlib import import_module
from typing import List, Mapping, Optional

from elasticsearch import Elasticsearch, ElasticsearchException, NotFoundError, RequestError, TransportError, helpers
from elasticsearch.serializer import JSONSerializer

from biothings.utils.common import inf, iter_n, merge, nan, splitstr, traverse

# the following imports used by utils.es.Database
from biothings.utils.dataload import dict_walk, update_dict_recur
from biothings.utils.dotfield import parse_dot_fields
from biothings.utils.hub_db import IDatabase
from biothings.utils.serializer import to_json

formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
es_logger = logging.getLogger("elasticsearch")
es_logger.setLevel(logging.WARNING)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
es_logger.addHandler(ch)

es_tracer = logging.getLogger("elasticsearch.trace")
es_tracer.setLevel(logging.WARNING)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
es_tracer.addHandler(ch)


def verify_ids(
    doc_iter,
    es_host,
    index,
    doc_type=None,
    step=100000,
):
    """verify how many docs from input interator/list overlapping with existing docs."""

    index = index
    doc_type = doc_type
    es = get_es(es_host)
    q = {"query": {"ids": {"values": []}}}
    total_cnt = 0
    found_cnt = 0
    out = []
    for doc_batch in iter_n(doc_iter, n=step):
        id_li = [doc["_id"] for doc in doc_batch]
        # id_li = [doc['_id'].replace('chr', '') for doc in doc_batch]
        q["query"]["ids"]["values"] = id_li
        xres = es.search(index=index, doc_type=doc_type, body=q, _source=False)
        found_cnt += xres["hits"]["total"]
        total_cnt += len(id_li)
        out.extend([x["_id"] for x in xres["hits"]["hits"]])
    return out


def get_es(es_host, timeout=120, max_retries=3, retry_on_timeout=False):
    es = Elasticsearch(es_host, timeout=timeout, max_retries=max_retries, retry_on_timeout=retry_on_timeout)
    return es


# WARNING: this wrapper changes the _index and _doc_type
#  usually this is unwanted but we will leave it this way here
def wrapper(func):
    """this wrapper allows passing index and doc_type from wrapped method."""

    def outter_fn(*args, **kwargs):
        self = args[0]
        index = kwargs.pop("index", self._index)  # pylint: disable=protected-access
        doc_type = kwargs.pop("doc_type", self._doc_type)  # pylint: disable=protected-access
        self._index = index  # pylint: disable=protected-access
        self._doc_type = doc_type  # pylint: disable=protected-access
        return func(*args, **kwargs)

    outter_fn.__doc__ = func.__doc__
    return outter_fn


class IndexerException(Exception):
    pass


@functools.lru_cache()
def get_doc_type(es_client, index_name):
    if int(es_client.info()["version"]["number"].split(".")[0]) < 7:
        mappings = es_client.indices.get_mapping(index_name)
        mappings = mappings[index_name]["mappings"]
        return next(iter(mappings.keys()))
    return None


class ESIndex:
    """An Elasticsearch Index Wrapping A Client.
    Counterpart for pymongo.collection.Collection"""

    # a new approach to biothings.utils.es.ESIndexer
    # but not intended to be a replacement in features.

    def __init__(self, client, index_name):
        self.client = client  # synchronous
        self.index_name = index_name  # MUST exist

    @property
    def doc_type(self):
        return get_doc_type(self.client, self.index_name)
        # TODO: remove the code block below if the above works
        # if int(self.client.info()["version"]["number"].split(".")[0]) < 7:
        #     mappings = self.client.indices.get_mapping(self.index_name)
        #     mappings = mappings[self.index_name]["mappings"]
        #     return next(iter(mappings.keys()))
        # return None

    # SUBCLASS NOTE &&
    # BEFORE YOU ADD A METHOD UNDER THIS CLASS:

    # An object of this class refers to an existing ES index. All operations
    # should target this index. Do not put uncommon methods here. They belong
    # to the subclasses. This class is to provide a common framework to support
    # Index-specific ES operations, an concept does not exist in the low-level
    # ES library, thus only providing low-level common operations, like doc_type
    # parsing across ES versions for biothings usecases. (single type per index)


class ESIndexer:
    # RETIRING THIS CLASS

    # --
    # Since we don't always directly work on
    # existing actual indices, the index referred here
    # can be an alias or does not exist.

    def __init__(
        self,
        index,
        doc_type="_doc",
        es_host="localhost:9200",
        step=500,
        step_size=10,  # elasticsearch.helpers.bulk
        number_of_shards=1,
        number_of_replicas=0,
        check_index=True,
        **kwargs,
    ):
        self.es_host = es_host
        self._es = get_es(es_host, **kwargs)
        self._host_major_ver = int(self._es.info()["version"]["number"].split(".")[0])
        # the name of the index when ESIndexer is initialized
        self.canonical_index_name = index
        self._index = index  # placeholder, will be updated later
        if check_index:
            # if index is actually an alias, resolve the alias to
            # the real underlying index
            self.check_index()

        self._doc_type = None
        if doc_type:
            self._doc_type = doc_type
        else:
            # assuming index exists, get mapping to discover doc_type
            try:
                m = self.get_mapping()
                assert len(m) == 1, "Expected only one doc type, got: %s" % m.keys()
                self._doc_type = list(m).pop()
            except Exception as e:  # pylint: disable=broad-except
                if check_index:
                    logging.info("Failed to guess doc_type: %s", e)
        self.number_of_shards = number_of_shards  # set number_of_shards when create_index
        self.number_of_replicas = int(number_of_replicas)  # set number_of_replicas when create_index
        self.step = step or 500  # the bulk size when doing bulk operation.
        self.step_size = (step_size or 10) * 1048576  # MB -> bytes
        self.s = None  # number of records to skip, useful to continue indexing after an error.

    def check_index(self):
        """
        Check if index is an alias, and update self._index to point to
        actual index

        TODO: the overall design of ESIndexer is not great. If we are exposing ES
         implementation details (such as the abilities to create and delete indices,
         create and update aliases, etc.) to the user of this Class, then this method
         doesn't seem that out of place.
        """
        try:
            res = self._es.indices.get_alias(name=self.canonical_index_name)
            # this is an alias
            if len(res) != 1:
                raise RuntimeError(f"Alias '{self.canonical_index_name}' does not associate with exactly 1 index")
            self._index = list(res.keys())[0]
        except NotFoundError:
            # probably intended to be an index name
            self._index = self.canonical_index_name

    @wrapper
    def get_biothing(self, bid, only_source=False, **kwargs):
        rawdoc = self._es.get(index=self._index, id=bid, doc_type=self._doc_type, **kwargs)
        if not only_source:
            return rawdoc
        else:
            doc = rawdoc["_source"]
            doc["_id"] = rawdoc["_id"]
            return doc

    @wrapper
    def exists(self, bid):
        """return True/False if a biothing id exists or not."""
        try:
            doc = self.get_biothing(bid, stored_fields=None)
            return doc["found"]
        except NotFoundError:
            return False

    @wrapper
    def mexists(self, bid_list):
        q = {"query": {"ids": {"values": bid_list}}}
        res = self._es.search(
            index=self._index,
            doc_type=self._doc_type,
            body=q,
            stored_fields=None,
            size=len(bid_list),
        )
        # id_set = set([doc['_id'] for doc in res['hits']['hits']])     # TODO: Confirm this line
        id_set = {doc["_id"] for doc in res["hits"]["hits"]}
        return [(bid, bid in id_set) for bid in bid_list]

    @wrapper
    def count(self, q=None, raw=False):
        try:
            count_kwargs = {"index": self._index}
            if q is not None:
                count_kwargs.update({"doc_type": self._doc_type, "q": q})
            _res = self._es.count(**count_kwargs)
            return _res if raw else _res["count"]
        except NotFoundError:
            return None

    @wrapper
    def count_src(self, src):
        if isinstance(src, str):
            src = [src]
        cnt_d = {}
        for _src in src:
            q = {"query": {"constant_score": {"filter": {"exists": {"field": _src}}}}}
            cnt_d[_src] = self.count(q)
        return cnt_d

    @wrapper
    def create_index(self, mapping=None, extra_settings=None):
        if not self._es.indices.exists(index=self._index):
            body = {
                "settings": {
                    "number_of_shards": self.number_of_shards,
                    "number_of_replicas": self.number_of_replicas,
                }
            }
            extra_settings = extra_settings or {}
            self.sanitize_settings(extra_settings)
            body["settings"].update(extra_settings)
            if mapping:
                # the mapping is passed in for elasticsearch 6
                # if the remote server is of elasticsearch version 7 or later
                # drop the doc_type first level key as it is no longer supported
                self._populate_es_version()
                if self._host_major_ver > 6:
                    if len(mapping) == 1 and next(iter(mapping)) not in (
                        "properties",
                        "dynamic",
                        "_meta",
                    ):
                        mapping = next(iter(mapping.values()))

                mapping = {"mappings": mapping}
                body.update(mapping)
            self._es.indices.create(index=self._index, **body)

    def _populate_es_version(self):
        if not hasattr(self, "_es_version"):
            self._es_version = int(self._es.info()["version"]["number"].split(".")[0])

    def exists_index(self, index: Optional[str] = None):
        if not index:
            index = self._index
        return self._es.indices.exists(index)

    def index(self, doc, id=None, action="index"):  # pylint: disable=redefined-builtin
        """add a doc to the index. If id is not None, the existing doc will be
        updated.
        """
        self._es.index(index=self._index, doc_type=self._doc_type, body=doc, id=id, params={"op_type": action})

    def index_bulk(self, docs, step=None, action="index"):
        self._populate_es_version()
        index_name = self._index
        doc_type = self._doc_type
        step = step or self.step

        def _get_bulk(doc):
            # keep original doc
            ndoc = copy.copy(doc)
            ndoc.update(
                {
                    "_index": index_name,
                    "_type": doc_type,
                    "_op_type": action,
                }
            )
            if self._host_major_ver > 6:
                ndoc.pop("_type")
            return ndoc

        actions = (_get_bulk(doc) for doc in docs)
        num_ok, errors = helpers.bulk(self._es, actions, chunk_size=step, max_chunk_bytes=self.step_size)
        if errors:
            raise ElasticsearchException("%d errors while bulk-indexing: %s" % (len(errors), [str(e) for e in errors]))
        return num_ok, errors

    def delete_doc(self, id):  # pylint: disable=redefined-builtin
        """delete a doc from the index based on passed id."""
        return self._es.delete(index=self._index, doc_type=self._doc_type, id=id)

    def delete_docs(self, ids, step=None):
        """delete a list of docs in bulk."""
        index_name = self._index
        doc_type = self._doc_type
        step = step or self.step

        def _get_bulk(_id):
            if self._host_major_ver >= 7:
                doc = {"_op_type": "delete", "_index": index_name, "_id": _id}
            else:
                doc = {"_op_type": "delete", "_index": index_name, "_type": doc_type, "_id": _id}
            return doc

        actions = (_get_bulk(_id) for _id in ids)
        return helpers.bulk(self._es, actions, chunk_size=step, stats_only=True, raise_on_error=False)

    def delete_index(self, index=None):
        if not index:
            index = self._index
        self._es.indices.delete(index)

    def update(self, id, extra_doc, upsert=True):  # pylint: disable=redefined-builtin
        """update an existing doc with extra_doc.
        allow to set upsert=True, to insert new docs.
        """
        body = {"doc": extra_doc}
        if upsert:
            body["doc_as_upsert"] = True
        return self._es.update(index=self._index, doc_type=self._doc_type, id=id, body=body)

    def update_docs(self, partial_docs, upsert=True, step=None, **kwargs):
        """update a list of partial_docs in bulk.
        allow to set upsert=True, to insert new docs.
        """
        index_name = self._index
        doc_type = self._doc_type
        step = step or self.step

        def _get_bulk(doc):
            if self._host_major_ver >= 7:
                doc = {"_op_type": "update", "_index": index_name, "_id": doc["_id"], "doc": doc}
            else:
                doc = {
                    "_op_type": "update",
                    "_index": index_name,
                    "_type": doc_type,
                    "_id": doc["_id"],
                    "doc": doc,
                }
            if upsert:
                doc["doc_as_upsert"] = True
            return doc

        actions = (_get_bulk(doc) for doc in partial_docs)
        return helpers.bulk(self._es, actions, chunk_size=step, **kwargs)

    def get_mapping(self):
        """return the current index mapping"""
        if self._host_major_ver <= 6:
            m = self._es.indices.get_mapping(
                index=self._index,
                doc_type=self._doc_type,
            )
            return m[self._index]["mappings"]
        elif self._host_major_ver <= 8:
            m = self._es.indices.get_mapping(index=self._index)
            # fake the mapping doc_type
            m = {self._doc_type: m[self._index]["mappings"]}
            return m
        else:
            raise RuntimeError(
                f"Server Elasticsearch version is {self._host_major_ver} "
                "which is unsupported when using old ESIndexer class"
            )

    def update_mapping(self, m):
        if self._host_major_ver <= 6:
            assert list(m) == [self._doc_type], "Bad mapping format, should have one doc_type, got: %s" % list(m)
            assert "properties" in m[self._doc_type], "Bad mapping format, no 'properties' key"
            return self._es.indices.put_mapping(index=self._index, doc_type=self._doc_type, body=m)
        elif self._host_major_ver <= 8:
            # this is basically guessing based on heuristics
            if len(m) == 1:
                if "properties" not in m:  # basically {'_doc': mapping}
                    m = next(iter(m.values()))  # take out the mapping
                else:  # basically just the mapping or type: properties
                    pass  # ignoring the possibility of doc_type='properties'
            else:  # we don't expect {} as input, so len(m) > 1
                if "properties" not in m:
                    raise RuntimeError(f"Possibly invalid mapping {m}")
            return self._es.indices.put_mapping(index=self._index, body=m)
        else:
            raise RuntimeError(
                f"Server Elasticsearch version is {self._host_major_ver} "
                "which is unsupported when using old ESIndexer class"
            )

    def get_mapping_meta(self):
        """return the current _meta field."""
        m = self.get_mapping()
        doc_type = self._doc_type
        if doc_type is None:
            # fetch doc_type from mapping

            assert len(m) == 1, (
                "More than one doc_type found, not supported when self._doc_type " + "is not initialized"
            )
            doc_type = list(m.keys())[0]
        return {"_meta": m[doc_type]["_meta"]}

    def update_mapping_meta(self, meta):
        allowed_keys = {"_meta", "_timestamp"}
        # if isinstance(meta, dict) and len(set(meta) - allowed_keys) == 0:
        if isinstance(meta, dict) and not set(meta) - allowed_keys:
            if self._host_major_ver >= 7:
                return self._es.indices.put_mapping(
                    index=self._index,
                    body=meta,
                )
            else:  # not sure if _type needs to be specified
                body = {self._doc_type: meta}
                return self._es.indices.put_mapping(doc_type=self._doc_type, body=body, index=self._index)
        else:
            raise ValueError('Input "meta" should have and only have "_meta" field.')

    def flush_and_refresh(self):
        try:
            self._es.indices.flush()
            self._es.indices.refresh()
        except:  # pylint: disable=bare-except  # noqa
            pass

    @wrapper
    def build_index(self, collection, verbose=True, query=None, bulk=True, update=False, allow_upsert=True):
        # update some settings for bulk indexing
        body = {
            "index": {
                "refresh_interval": "-1",  # disable refresh temporarily
                "auto_expand_replicas": "0-all",
            }
        }
        self.update_settings(body)
        try:
            self._build_index_sequential(collection, verbose, query=query, bulk=bulk, update=update, allow_upsert=True)
        finally:
            # restore some settings after bulk indexing is done.
            body = {"index": {"refresh_interval": "1s"}}  # default settings
            self.update_settings(body)

            self.flush_and_refresh()

            time.sleep(1)
            src_cnt = collection.count(query)
            es_cnt = self.count()
            if src_cnt != es_cnt:
                raise IndexerException(
                    "Total count of documents does not match [{}, should be {}]".format(es_cnt, src_cnt)
                )

            return es_cnt  # NOQA B012

    def _build_index_sequential(
        self, collection, verbose=False, query=None, bulk=True, update=False, allow_upsert=True
    ):
        def rate_control(cnt, t):
            delay = 0
            if t > 90:
                delay = 30
            elif t > 60:
                delay = 10
            if delay:
                time.sleep(delay)

        from biothings.utils.mongo import doc_feeder

        src_docs = doc_feeder(collection, step=self.step, s=self.s, batch_callback=rate_control, query=query)
        if bulk:
            if update:
                # input doc will update existing one
                # if allow_upsert, create new one if not exist
                res = self.update_docs(src_docs, upsert=allow_upsert)
            else:
                # input doc will overwrite existing one
                res = self.index_bulk(src_docs)
            # if len(res[1]) > 0:
            if res[1]:
                raise IndexerException("Error: {} docs failed indexing.".format(len(res[1])))
            return res[0]

        else:
            cnt = 0
            for doc in src_docs:
                self.index(doc)
                cnt += 1
            return cnt

    @wrapper
    def optimize(self, max_num_segments=1):
        """optimize the default index."""
        params = {
            "wait_for_merge": False,
            "max_num_segments": max_num_segments,
        }
        return self._es.indices.forcemerge(index=self._index, params=params)

    def clean_field(self, field, dryrun=True, step=5000):
        """remove a top-level field from ES index, if the field is the only field of the doc,
        remove the doc as well.
        step is the size of bulk update on ES
        try first with dryrun turned on, and then perform the actual updates with dryrun off.
        """
        if self._host_major_ver >= 7:
            raise RuntimeError("clean_field is no longer supported")
        q = {"query": {"constant_score": {"filter": {"exists": {"field": field}}}}}
        cnt_orphan_doc = 0
        cnt = 0
        _li = []
        for doc in self.doc_feeder(query=q):
            if set(doc) == {"_id", field}:
                cnt_orphan_doc += 1
                # delete orphan doc
                _li.append({"delete": {"_index": self._index, "_type": self._doc_type, "_id": doc["_id"]}})
            else:
                # otherwise, just remove the field from the doc
                _li.append({"update": {"_index": self._index, "_type": self._doc_type, "_id": doc["_id"]}})
                # this script update requires "script.disable_dynamic: false" setting
                # in elasticsearch.yml
                _li.append({"script": 'ctx._source.remove("{}")'.format(field)})

            cnt += 1
            if len(_li) == step:
                if not dryrun:
                    self._es.bulk(body=_li)
                _li = []
        if _li:
            if not dryrun:
                self._es.bulk(body=_li)

        return {"total": cnt, "updated": cnt - cnt_orphan_doc, "deleted": cnt_orphan_doc}

    @wrapper
    def doc_feeder_using_helper(self, step=None, verbose=True, query=None, scroll="10m", **kwargs):
        # verbose unimplemented
        step = step or self.step
        q = query if query else {"query": {"match_all": {}}}
        for rawdoc in helpers.scan(
            client=self._es,
            query=q,
            scroll=scroll,
            index=self._index,
            doc_type=self._doc_type,
            **kwargs,
        ):
            if rawdoc.get("_source", False):
                doc = rawdoc["_source"]
                doc["_id"] = rawdoc["_id"]
                yield doc
            else:
                yield rawdoc

    @wrapper
    def doc_feeder(self, step=None, verbose=True, query=None, scroll="10m", only_source=True, **kwargs):
        step = step or self.step
        q = query if query else {"query": {"match_all": {}}}
        _q_cnt = self.count(q=q, raw=True)
        n = _q_cnt["count"]
        n_shards = _q_cnt["_shards"]["total"]
        assert n_shards == _q_cnt["_shards"]["successful"]
        # Not sure if scroll size is per shard anymore in the new ES...should check this
        _size = int(step / n_shards)
        assert _size * n_shards == step
        cnt = 0
        # t0 = time.time()
        # if verbose:
        #     t1 = time.time()

        res = self._es.search(
            index=self._index,
            doc_type=self._doc_type,
            body=q,
            size=_size,
            search_type="scan",
            scroll=scroll,
            **kwargs,
        )
        # double check initial scroll request returns no hits
        # assert len(res['hits']['hits']) == 0
        assert not res["hits"]["hits"]

        while True:
            # if verbose:
            #     t1 = time.time()
            res = self._es.scroll(scroll_id=res["_scroll_id"], scroll=scroll)
            # if len(res['hits']['hits']) == 0:
            if not res["hits"]["hits"]:
                break
            else:
                for rawdoc in res["hits"]["hits"]:
                    if rawdoc.get("_source", False) and only_source:
                        doc = rawdoc["_source"]
                        doc["_id"] = rawdoc["_id"]
                        yield doc
                    else:
                        yield rawdoc
                    cnt += 1

        assert cnt == n, "Error: scroll query terminated early [{}, {}], please retry.\nLast response:\n{}".format(
            cnt, n, res
        )

    @wrapper
    def get_id_list(self, step=None, verbose=True):
        step = step or self.step
        cur = self.doc_feeder(step=step, _source=False, verbose=verbose)
        for doc in cur:
            yield doc["_id"]

    @wrapper
    def get_docs(self, ids, step=None, only_source=True, **mget_args):
        """Return matching docs for given ids iterable, if not found return None.
        A generator is returned to the matched docs.  If only_source is False,
        the entire document is returned, otherwise only the source is returned."""
        # chunkify
        step = step or self.step
        for chunk in iter_n(ids, step):
            if self._host_major_ver > 6:
                chunk_res = self._es.mget(body={"ids": chunk}, index=self._index, **mget_args)
            else:
                chunk_res = self._es.mget(body={"ids": chunk}, index=self._index, doc_type=self._doc_type, **mget_args)
            for rawdoc in chunk_res["docs"]:
                if ("found" not in rawdoc) or (("found" in rawdoc) and not rawdoc["found"]):
                    continue
                elif not only_source:
                    yield rawdoc
                else:
                    doc = rawdoc["_source"]
                    doc["_id"] = rawdoc["_id"]
                    yield doc

    def find_biggest_doc(self, fields_li, min=5, return_doc=False):  # pylint: disable=redefined-builtin
        """return the doc with the max number of fields from fields_li."""
        for n in range(len(fields_li), min - 1, -1):
            for field_set in itertools.combinations(fields_li, n):
                q = " AND ".join(["_exists_:" + field for field in field_set])
                q = {"query": {"query_string": {"query": q}}}
                cnt = self.count(q)
                if cnt > 0:
                    if return_doc:
                        res = self._es.search(index=self._index, doc_type=self._doc_type, body=q, size=cnt)
                        return res
                    else:
                        return (cnt, q)

    def snapshot(self, repo, snapshot, mode=None, **params):
        body = {
            "indices": self._index,
            "include_global_state": False
            # there is no reason to include global state in our application
            # we want to separate the staging env from the production env
            # (global state includes index templates and ingest pipeline)
            # but this doesn't mean this setting has to be here
            # maybe move this line to where it belongs later
        }
        if mode == "purge":
            # Note: this works, just for small one when deletion is done instantly
            try:
                self._es.snapshot.get(repo, snapshot)
                # if we can get it, we have to delete it
                self._es.snapshot.delete(repo, snapshot)
            except NotFoundError:
                # ok, nothing to delete/purge
                pass
        try:
            return self._es.snapshot.create(repo, snapshot, body=body, params=params)
        except RequestError as e:
            try:
                err_msg = e.info["error"]["reason"]
            except KeyError:
                err_msg = e.error
            raise IndexerException("Can't snapshot '%s': %s" % (self._index, err_msg))

    def restore(self, repo_name, snapshot_name, index_name=None, purge=False):
        indices = self.get_indices_from_snapshots(repo_name, snapshot_name)
        if not index_name:
            index_name = indices[0]
        indices = ",".join(indices)

        if purge:
            try:
                self._es.indices.get(index=index_name)
                # if we get there, it exists, delete it
                self._es.indices.delete(index=index_name)
            except NotFoundError:
                # no need to delete it,
                pass
        try:
            # this is just about renaming index within snapshot to index_name
            body = {
                "indices": indices,
                "rename_replacement": index_name,
                "ignore_unavailable": True,
                "rename_pattern": "(.+)",
                # set to False, snapshots were created without global state anyway.
                #  In ES8, an error will be raised if set to True
                "include_global_state": False,
            }
            return self._es.snapshot.restore(repo_name, snapshot_name, body=body)
        except TransportError as e:
            raise IndexerException(
                "Can't restore snapshot '%s' (does index '%s' already exist ?): %s" % (snapshot_name, index_name, e)
            )

    def get_alias(self, index: str = None, alias_name: str = None) -> List[str]:
        """
        Get indices with alias associated with given index name or alias name

        Args:
            index: name of index
            alias_name: name of alias

        Returns:
            Mapping of index names with their aliases
        """
        return self._es.indices.get_alias(index=index, name=alias_name)

    def get_settings(self, index: str = None) -> Mapping[str, Mapping]:
        """
        Get indices with settings associated with given index name

        Args:
            index: name of index

        Returns:
            Mapping of index names with their settings
        """
        return self._es.indices.get_settings(index=index)

    def get_indice_names_by_settings(self, index: str = None, sort_by_creation_date=False, reverse=False) -> List[str]:
        """
        Get list of indices names associated with given index name, using indices' settings

        Args:
            index: name of index
            sort_by_creation_date: sort the result by indice's creation_date
            reverse: control the direction of the sorting

        Returns:
            list of index names (str)
        """
        indices_settings = self.get_settings(index)
        names_with_creation_date = [
            (indice_name, setting["settings"]["index"]["creation_date"])
            for indice_name, setting in indices_settings.items()
        ]

        if sort_by_creation_date:
            names_with_creation_date = sorted(
                names_with_creation_date,
                key=lambda name_with_creation_date: name_with_creation_date[1],
                reverse=reverse,
            )

        indice_names = [name for name, _ in names_with_creation_date]
        return indice_names

    def update_alias(self, alias_name: str, index: Optional[str] = None):
        """
        Create or update an ES alias pointing to an index

        Creates or updates an alias in Elasticsearch, associated with the
        given index name or the underlying index of the ESIndexer instance.

        When the alias name does not exist, it will be created. If an existing
        alias already exists, it will be updated to only associate with the
        index.

        When the alias name already exists, an exception will be raised,
        UNLESS the alias name is the same as index name that the ESIndexer is
        initialized with. In this case, the existing index with the name
        collision will be deleted, and the alias will be created
        in its place. This feature is intended for seamless migration from
        an index to an alias associated with an index for zero-downtime
        installs.

        Args:
            alias_name: name of the alias
            index: name of the index to associate with alias. If None, the
                index of the ESIndexer instance is used.

        Raises:
            IndexerException
        """
        if index is None:
            index = self._index

        if not self._es.indices.exists(alias_name):
            # case 1 it does not already exist
            #  create the alias pointing to _index
            self._es.indices.put_alias(index=index, name=alias_name)
        else:  # case 2 it already exists
            if self._es.indices.exists_alias(name=alias_name):
                # if it is an alias, blindly update
                #  This removes any other indices associated with the alias
                actions = {"actions": [{"add": {"index": index, "alias": alias_name}}]}
                removes = [
                    {"remove": {"index": index_name, "alias": alias_name}} for index_name in self.get_alias(alias_name)
                ]
                actions["actions"].extend(removes)
                self._es.indices.update_aliases(body=actions)
            else:  # it is an index
                # if not _index and is the canonical index name
                #  then delete the index and create alias
                if alias_name == self.canonical_index_name:
                    self._es.indices.delete(alias_name)
                    self._es.indices.put_alias(index=index, name=alias_name)
                else:
                    raise IndexerException(
                        f"Cannot create alias {alias_name} " "an index with the same name " "already exists"
                    )

    def get_repository(self, repo_name):
        try:
            return self._es.snapshot.get_repository(repo_name)
        except NotFoundError:
            raise IndexerException("Repository '%s' doesn't exist" % repo_name)

    def create_repository(self, repo_name, settings):
        try:
            self._es.snapshot.create_repository(repo_name, settings)
        except TransportError as e:
            raise IndexerException("Can't create snapshot repository '%s': %s" % (repo_name, e))

    def get_snapshots(self, repo_name, snapshot_name):
        try:
            snapshots = self._es.snapshot.get(repository=repo_name, snapshot=snapshot_name)
            return [snapshot for snapshot in snapshots["snapshots"]]
        except NotFoundError as ex:
            message = str(ex)
            if ex.error == "repository_missing_exception":
                message = "Repository '%s' doesn't exist" % repo_name
            if ex.error == "snapshot_missing_exception":
                message = "Snapshot '%s' doesn't exist" % snapshot_name
            raise IndexerException(message)

    def get_indices_from_snapshots(self, repo_name, snapshot_name):
        snapshots = self.get_snapshots(repo_name, snapshot_name)
        indices = []
        for snapshot in snapshots:
            indices.extend([index for index in (snapshot.get("indices") or [])])
        return indices

    def get_snapshot_status(self, repo, snapshot):
        return self._es.snapshot.status(repo, snapshot)

    def get_restore_status(self, index_name=None):
        index_name = index_name or self._index
        recov = self._es.indices.recovery(index=index_name)
        if index_name not in recov:
            return {"status": "INIT", "progress": "0%"}
        shards = recov[index_name]["shards"]
        # get all shards status
        shards_status = [s["stage"] for s in shards]
        done = len([s for s in shards_status if s == "DONE"])
        if set(shards_status) == {"DONE"}:
            return {"status": "DONE", "progress": "100%"}
        else:
            return {
                "status": "IN_PROGRESS",
                "progress": "%.2f%%" % (done / len(shards_status) * 100),
            }

    def get_internal_number_of_replicas(self):
        try:
            index_settings = self._es.indices.get_settings(self._index)
            return index_settings[self._index]["settings"]["index"]["number_of_replicas"]
        except Exception:
            return

    def set_internal_number_of_replicas(self, number_of_replicas=None):
        if not number_of_replicas:
            number_of_replicas = self.number_of_replicas
        settings = {"index": {"number_of_replicas": number_of_replicas}}
        self.update_settings(settings)

    def sanitize_settings(self, settings):
        """
        Clean up settings dictinary to remove those static fields cannot be updated.
         like: "uuid", "provided_name", "creation_date", "version",
        settings will be updated  both in-place and returned as well.
        """
        setting_fields_to_remove = [
            "uuid",
            "provided_name",
            "creation_date",
            "version",
            "number_of_shards",
        ]
        for field in setting_fields_to_remove:
            settings["index"].pop(field, None)
        return settings

    def update_settings(self, settings, close=False, **params):
        """
        Parameters:
            - settings: should be valid ES index's settings.
            - close: In order to update static settings, the index must be closed first.

        Ref: https://www.elastic.co/guide/en/elasticsearch/reference/7.17/index-modules.html#index-modules-settings
        """
        # Some static fields like: "uuid", "provided_name", "creation_date", "version",
        #   "number_of_shards", which ES doesn't allow update them.
        self.sanitize_settings(settings)

        if close:
            self.close()

        self._es.indices.put_settings(
            body=settings,
            index=self._index,
            params=params,
        )

        if close:
            self.open()

    def reindex(self, src_index, is_remote=False, **kwargs):
        """In order to reindex from remote,
        - src es_host must be set to an IP which the current ES host can connect to.
            It means that if 2 indices locate in same host, the es_host can be set to localhost,
            but if they are in different hosts, an IP must be used instead.
        - If src host uses SSL, https must be included in es_host. Eg: https://192.168.1.10:9200
        - src host must be whitelisted in the current ES host.
            Ref: https://www.elastic.co/guide/en/elasticsearch/reference/7.17/reindex-upgrade-remote.html
        """
        body = {
            "source": {"index": src_index._index},
            "dest": {"index": self._index},
        }
        if is_remote:
            host = src_index.es_host
            if not host.startswith("http"):
                host = "http://" + host
            body["source"]["remote"] = {"host": host}
        return self._es.reindex(body=body, **kwargs)

    def close(self):
        self._es.indices.close(self._index)

    def open(self):
        self._es.indices.open(self._index)


class MappingError(Exception):
    pass


def generate_es_mapping(inspect_doc, init=True, level=0):
    """Generate an ES mapping according to "inspect_doc", which is
    produced by biothings.utils.inspect module"""
    map_tpl = {
        int: {"type": "integer"},
        bool: {"type": "boolean"},
        float: {"type": "float"},
        str: {
            "type": "keyword",
            "normalizer": "keyword_lowercase_normalizer",
        },  # not splittable (like an ID for instance)
        splitstr: {"type": "text"},
    }
    # inspect_doc, if it's been jsonified, contains keys with type as string,
    # such as "<class 'str'>". This is not a real type and we need to convert them
    # back to actual types. This is transparent if inspect_doc isalready in proper format
    pat = re.compile(r"<class '(\w+)'>")

    def str2type(k):
        if isinstance(k, str):
            mat = pat.findall(k)
            if mat:
                return eval(mat[0])  # actual type
            else:
                return k
        else:
            return k

    inspect_doc = dict_walk(inspect_doc, str2type)

    mapping = {}
    errors = []
    none_type = type(None)
    if init and "_id" not in inspect_doc:
        errors.append("No _id key found, document won't be indexed. (doc: %s)" % inspect_doc)
    for rootk in inspect_doc:
        if rootk == "_id":
            keys = list(inspect_doc[rootk].keys())
            if str in keys and splitstr in keys:
                keys.remove(str)
            if not len(keys) == 1 or (keys[0] != str and keys[0] != splitstr):
                errors.append("_id fields should all be a string type (got: %s)" % keys)
            # it was just a check, it's not part of the mapping
            continue
        if rootk == "_stats":
            continue
        if isinstance(rootk, type(None)):  # if rootk == type(None):
            # value can be null, just skip it
            continue
        # some inspect report have True as value, others have dict (will all have dict eventually)
        if inspect_doc[rootk] is True:
            inspect_doc[rootk] = {}
        keys = list(inspect_doc[rootk].keys())
        # if dict, it can be a dict containing the type (no explore needed) or a dict
        # containing more keys (explore needed)
        if list in keys:
            # we explore directly the list w/ inspect_doc[rootk][list] as param.
            # (similar to skipping list type, as there's no such list type in ES mapping)
            # carefull: there could be list of list, if which case we move further into the structure
            # to skip them
            toexplore = inspect_doc[rootk][list]
            while list in toexplore:
                toexplore = toexplore[list]
            if len(toexplore) > 1:
                # we want to make sure that, whatever the structure, the types involved were the same
                # Exception: None is allowed with other types (translates to 'null' in ES)
                # other_types = set([k for k in toexplore.keys() if k != list and isinstance(k, type) and k is not type(None)])    # TODO: Confirm this line
                other_types = {
                    k for k in toexplore.keys() if k != list and isinstance(k, type) and not isinstance(k, none_type)
                }
                # some mixes are allowed by ES
                if {int, float}.issubset(other_types):
                    other_types.discard(int)  # float > int
                    toexplore.pop(int)
                if len(other_types) > 1:
                    raise Exception("Mixing types for key '%s': %s" % (rootk, other_types))
            res = generate_es_mapping(toexplore, init=False, level=level + 1)
            # is it the only key or do we have more ? (ie. some docs have data as "x", some
            # others have list("x")
            # list was either a list of values (end of tree) or a list of dict. Depending
            # on that, we add "properties" (when list of dict) or not (when list of values)
            if type in set(map(type, inspect_doc[rootk][list])):
                mapping[rootk] = res
            else:
                mapping[rootk] = {"properties": {}}
                mapping[rootk]["properties"] = res
        elif set(map(type, keys)) == {type}:
            # it's a type declaration, no explore
            # typs = list(map(type, [k for k in keys if k is not type(None)]))    # TODO: Confirm this line
            typs = list(map(type, [k for k in keys if not isinstance(k, none_type)]))
            if len(typs) > 1:
                errors.append("More than one type (key:%s,types:%s)" % (repr(rootk), repr(keys)))
            try:
                typ = list(inspect_doc[rootk].keys())
                # ther can still be more than one type, if we have a None combined with
                # the "correct" one. We allow None as a combined type, but we want to ignore
                # it when we want to find the mapping
                if len(typ) == 1:
                    typ = typ[0]
                else:
                    # typ = [t for t in typ if t is not type(None)][0]      # TODO: Confirm this line
                    typ = [t for t in typ if not isinstance(t, none_type)][0]
                if typ is nan or typ is inf:
                    raise TypeError(typ)
                mapping[rootk] = map_tpl[typ]
            except KeyError:
                errors.append("Can't find map type %s for key %s" % (inspect_doc[rootk], rootk))
            except TypeError:
                errors.append("Type %s for key %s isn't allowed in ES mapping" % (typ, rootk))
        elif inspect_doc[rootk] == {}:
            typ = rootk
            return map_tpl[typ]
        else:
            mapping[rootk] = {"properties": {}}
            mapping[rootk]["properties"] = generate_es_mapping(inspect_doc[rootk], init=False, level=level + 1)
    if errors:
        raise MappingError("Error while generating mapping", errors)
    return mapping


# from biothings.utils.common import json_serial
# from biothings.utils.dataload import update_dict_recur
# from biothings.utils.dotfield import parse_dot_fields

########################
# ES as HUB DB backend #
########################
# from biothings.utils.hub_db import IDatabase


def get_hub_db_conn():
    return Database()


def get_src_conn():
    return get_hub_db_conn()


def get_src_db(conn=None):
    db = Database()
    return db[db.CONFIG.DATA_SRC_DATABASE]


def get_src_dump():
    db = Database()
    return db[db.CONFIG.DATA_SRC_DUMP_COLLECTION]


def get_src_master():
    db = Database()
    return db[db.CONFIG.DATA_SRC_MASTER_COLLECTION]


def get_src_build():
    db = Database()
    return db[db.CONFIG.DATA_SRC_BUILD_COLLECTION]


def get_src_build_config():
    db = Database()
    return db[db.CONFIG.DATA_SRC_BUILD_CONFIG_COLLECTION]


def get_data_plugin():
    db = Database()
    return db[db.CONFIG.DATA_PLUGIN_COLLECTION]


def get_api():
    db = Database()
    return db[db.CONFIG.API_COLLECTION]


def get_cmd():
    db = Database()
    return db[db.CONFIG.CMD_COLLECTION]


def get_event():
    db = Database()
    return db[db.CONFIG.EVENT_COLLECTION]


def get_hub_config():
    db = Database()
    return db[getattr(db.CONFIG, "HUB_CONFIG_COLLECTION", "hub_config")]


def get_source_fullname(col_name):
    return col_name


def get_last_command():
    cmds = list(sorted(get_cmd()._read().values(), key=lambda cmd: cmd["_id"]))
    return cmds[-1] if cmds else None


# ES7+ FOR HUB_DB
# IS *NOT* DESIGNED FOR
# MANAGING HEAVY WORKLOADS


class Database(IDatabase):
    CONFIG = None  # will be set by bt.utils.hub_db.setup()

    def __init__(self):
        super(Database, self).__init__()
        self.name = self.CONFIG.DATA_HUB_DB_DATABASE
        self.host = self.CONFIG.HUB_DB_BACKEND["host"]
        self.client = Elasticsearch(self.host, serializer=_HubDBEncoder())

        if not self.client.indices.exists(index=self.name):
            self.client.indices.create(
                index=self.name,
                settings={
                    "number_of_shards": 1,
                    "number_of_replicas": 0,
                },
                mappings={"enabled": False},
            )

    @property
    def address(self):
        return self.host

    # ES API OPS
    # ON "COLLECTION"

    def _exists(self, _id):
        return self.client.exists(index=self.name, id=_id)

    def _read(self, _id):
        doc = self.client.get(index=self.name, id=_id)
        return doc["_source"]

    def _write(self, _id, doc):
        assert doc.pop("_id", None) in (_id, None)
        self.client.index(index=self.name, document=doc, id=_id)
        self.client.indices.refresh(index=self.name)

    def _modify(self, _id, func):
        doc = self._read(_id)
        doc = func(doc) or doc
        self._write(_id, doc)

    # HIGH LEVEL
    # HUB_DB ABSTRACTION

    def create_collection(self, colname):
        return self[colname]

    def __getitem__(self, colname):
        return Collection(colname, self)


class Collection:
    def __init__(self, colname, db):
        self.name = colname
        self.db = db

        if not self.db._exists(colname):
            self._write({})

    # COLLECTION OPS

    def _read(self):
        return self.db._read(self.name)

    def _write(self, col):
        self.db._write(self.name, col)

    # COLLECTION DOC OPS

    def _exists_one(self, _id):
        return str(_id) in self._read()

    def _write_one(self, doc):
        def func(collection):
            collection[str(doc["_id"])] = doc

        self.db._modify(self.name, func)

    # HUB_DB ABSTRACTION

    def __getitem__(self, _id):
        return self.find_one({"_id": _id})

    def find_one(self, *args, **kwargs):
        results = self.find(*args, **kwargs)
        return results[0] if results else None

    def find(self, filter=None, projection=None, *args, **kwargs):
        if args or kwargs:
            raise NotImplementedError()

        results = []
        logger = logging.getLogger(__name__)
        for doc in self._read().values():
            _doc = dict(traverse(doc))  # dotdict
            _doc.update(dict(traverse(doc, True)))
            for k, v in (filter or {}).items():
                if isinstance(v, dict) and "$exists" in v:
                    logger.error("Ignored filter: {'%s': %s}", k, v)
                    continue
                if _doc.get(k) != v:
                    break
            else:  # no break
                results.append(_pyobj(doc))

        if projection:  # used by BuildManager.build_info
            logger.error("Ignored projection: %s", projection)

        return results

    def insert_one(self, document, *args, **kwargs):
        if args or kwargs:
            raise NotImplementedError()

        if self._exists_one(document["_id"]):
            raise ValueError("Document already exists.")

        self._write_one(document)

    def replace_one(self, filter, replacement, upsert=False, *args, **kwargs):
        if args or kwargs:
            raise NotImplementedError()

        doc = self.find_one(filter) or {}
        if not (doc or upsert):
            raise ValueError("No Match.")

        _id = doc.get("_id") or filter["_id"]
        replacement["_id"] = _id
        self._write_one(replacement)

    # update operations support
    # a subset of mongo operators
    # for partial document editing

    def _update_one(self, doc, update, *args, **kwargs):
        if args or kwargs:
            raise NotImplementedError()

        if not len(update) == 1:
            raise ValueError("Invalid operator.")

        if next(iter(update)) not in ("$set", "$unset", "$push", "$addToSet", "$pull"):
            raise NotImplementedError(next(iter(update)))

        # https://docs.mongodb.com/manual/reference/operator/update/set/
        # https://docs.mongodb.com/manual/reference/operator/update/unset/
        # https://docs.mongodb.com/manual/reference/operator/update/push/
        # https://docs.mongodb.com/manual/reference/operator/update/addToSet/
        # https://docs.mongodb.com/manual/reference/operator/update/pull/

        if "$set" in update:
            _update = json.loads(to_json(update["$set"]))
            _update = parse_dot_fields(_update)
            doc = update_dict_recur(doc, _update)

        elif "$unset" in update:
            for dotk, v in traverse(doc):
                if dotk in update["$unset"]:
                    v["__REMOVE__"] = True
            doc = merge({}, doc)

        elif "$push" in update:
            for key, val in update["$push"].items():
                if "." in key:  # not all mongo operators are fully implemented
                    raise NotImplementedError("nested key in $push: %s" % key)
                doc.setdefault(key, []).append(val)

        elif "$addToSet" in update:
            for key, val in update["$addToSet"].items():
                if "." in key:  # not all mongo operators are fully implemented
                    raise NotImplementedError("nested key in $addToSet: %s" % key)
                field = doc.setdefault(key, [])
                if val not in field:
                    field.append(val)

        else:  # "$pull" in update:
            for key, val in update["$pull"].items():
                if "." in key:  # not all mongo operators are fully implemented
                    raise NotImplementedError("nested key in $pull: %s" % key)
                if not isinstance(val, (str, int)):
                    raise NotImplementedError("value or condition in $pull: %s" % val)
                if isinstance(doc.get(key), list):
                    doc[key][:] = [x for x in doc[key] if x != val]

        self._write_one(doc)

    def update_one(self, filter, update, upsert=False, *args, **kwargs):
        doc = self.find_one(filter) or {}
        if not (doc or upsert):
            raise ValueError("No Match.")

        self._update_one(doc, update, *args, **kwargs)

    def update_many(self, filter, update, upsert=False, *args, **kwargs):
        docs = self.find(filter)
        if not docs and upsert:
            if any("." in k for k in filter):
                raise ValueError("dotfield in upsert.")
            docs = [filter]

        for doc in docs:
            self._update_one(doc, update, *args, **kwargs)

    # DEPRECATED
    # -----------------

    def update(self, *args, **kwargs):
        # In the future,
        # Use replace_one(), update_one(), or update_many() instead.
        self.update_many(*args, **kwargs)

    def save(self, doc, *args, **kwargs):
        # In the future,
        # Use insert_one() or replace_one() instead.
        self._write_one(doc)

    def remove(self, query):
        # In the future,
        # Use delete_one() or delete_many() instead.
        docs = self.find(query)
        collection = self._read()
        for doc in docs:
            del collection[doc["_id"]]
        self._write(collection)

    def count(self):
        # In the future,
        # Use count_documents() or estimated_document_count() instead.
        return len(self._read())


# JSON <-> BSON
# -----------------
# The original author of the biothings.hub decided to have the interface
# of hubdb to model that of MongoDB, the prevailing choice of hub db.
# However, MongoDB stores BSON, extended from JSON to add some optional
# non-JSON-native data types, like dates and binary data, while ES only
# supports JSON documents. Upper layers of biothings.hub has subsequently
# been developed to take full advantage of BSON and specifically have
# stored document values in datetime type in hubdb. Thus, the ES as hubdb
# feature needs to provide the support of these additional data types.

# To achieve generic support of additional data types, non-JSON-serializable
# data by default are serialized to their string representations using "repr".
# For example, datetime objects are stored like datetime.datetime(...).

# Is using "eval" safe? For our internal-use purpose, and combining the fact
# that this module is designed for small demonstrations, better genericity
# currently outweighs the potential security risk, that we may easily support
# serializing additional data types without manually handling each type.
# when the circumstances change, it is advised to reconsider this implementation.


def _pyobj(doc):  # ES doc -> Python object
    for _, _doc in traverse(doc):
        if isinstance(_doc, dict):
            for k, v in list(_doc.items()):
                _doc[k] = _eval(v)
        elif isinstance(_doc, list):
            _doc[:] = map(_eval, _doc)

    return doc


_PY_REPR = re.compile(r"([\w.]+)\(.*\)")


def _eval(v):
    try:
        match = _PY_REPR.match(v)
        if match:
            clsstr = match.group(1)
            modstr = clsstr.rsplit(".", 1)[0]
            return eval(v, {modstr: import_module(modstr)})
    except Exception:
        ...

    return v


class _HubDBEncoder(JSONSerializer):
    TYPES = (
        datetime.datetime,
        # ...
    )

    def default(self, obj):
        if isinstance(obj, self.TYPES):
            return repr(obj)
        return super().default(obj)



"""
Elasticsearch Handlers

biothings.web.handlers.BaseESRequestHandler

    Supports: (all features above and)
    - access to biothing_type attribute
    - access to ES query pipeline stages
    - pretty print elasticsearch exceptions
    - common control option out_format

    Subclasses:
    - biothings.web.handlers.MetadataSourceHandler
    - biothings.web.handlers.MetadataFieldHandler
    - myvariant.web.beacon.BeaconHandler

biothings.web.handlers.ESRequestHandler

    Supports: (all features above and)
    - common control options (raw, rawquery)
    - common transform options (dotfield, always_list...)
    - query pipeline customization hooks
    - single query through GET
    - multiple quers through POST

    Subclasses:
    - biothings.web.handlers.BiothingHandler
    - biothings.web.handlers.QueryHandler

"""

import logging
from collections import Counter
from inspect import iscoroutinefunction
from types import CoroutineType

from tornado.web import Finish, HTTPError

from biothings.utils import serializer
from biothings.web.analytics.events import GAEvent
from biothings.web.handlers.base import BaseAPIHandler
from biothings.web.query.pipeline import QueryPipelineException, QueryPipelineInterrupt

__all__ = [
    "BaseQueryHandler",
    "MetadataSourceHandler",
    "MetadataFieldHandler",
    "BiothingHandler",
    "QueryHandler",
]

logger = logging.getLogger(__name__)


class BaseQueryHandler(BaseAPIHandler):
    def initialize(self, biothing_type=None, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self.biothing_type = biothing_type
        self.pipeline = self.biothings.pipeline
        self.metadata = self.biothings.metadata

    def prepare(self):
        super().prepare()

        # provide convenient access to next stages
        self.args.biothing_type = self.biothing_type

        self.event = GAEvent(
            {
                "__secondary__": [],  # secondary analytical objective: field tracking
                "category": "{}_api".format(self.biothings.config.APP_VERSION),  # eg.'v1_api'
                "action": "_".join((self.name, self.request.method.lower())),  # eg.'query_get'
                # 'label': 'fetch_all', etc.
                # 'value': 100, # number of queries
            }
        )

        if self.args._source:
            fields = [str(field) for field in self.args._source]  # in case input is not str
            fields = [field.split(".", 1)[0] for field in fields]  # only consider root keys
            for field, cnt in Counter(fields).items():
                self.event["__secondary__"].append(
                    GAEvent({"category": "parameter_tracking", "action": "field_filter", "label": field, "value": cnt})
                )
        else:
            self.event["__secondary__"].append(
                GAEvent(
                    {
                        "category": "parameter_tracking",
                        "action": "field_filter",
                        "label": "all",
                    }
                )
            )

    def write(self, chunk):
        # add an additional header to the JSON formatter
        # with a header image and a title-like section
        # further more, show a link to documentation

        # relevant settings in config:
        # HTML_OUT_TITLE
        # HTML_OUT_HEADER_IMG
        # HTML_OUT_<ENDPOINT>_DOCS

        DEFAULT_TITLE = "<p>Biothings API</p>"
        DEFAULT_IMG = "https://biothings.io/static/favicon.ico"

        try:
            if self.format == "html":
                config = self.biothings.config
                chunk = self.render_string(
                    template_name="api.html",
                    data=serializer.to_json(chunk),
                    link=serializer.URL(self.request.full_url()).remove("format"),
                    title_div=getattr(config, "HTML_OUT_TITLE", "") or DEFAULT_TITLE,
                    header_img=getattr(config, "HTML_OUT_HEADER_IMG", "") or DEFAULT_IMG,
                    learn_more=getattr(config, f"HTML_OUT_{self.name.upper()}_DOCS", ""),
                )
                self.set_header("Content-Type", "text/html; charset=utf-8")
                return super(BaseAPIHandler, self).write(chunk)

        except Exception as exc:
            logger.warning(exc)

        super().write(chunk)


class MetadataSourceHandler(BaseQueryHandler):
    """
    GET /metadata
    """

    name = "metadata"
    kwargs = dict(BaseQueryHandler.kwargs)
    kwargs["GET"] = {
        "dev": {"type": bool, "default": False},
        "raw": {"type": bool, "default": False},
    }

    async def get(self):
        info = await self.metadata.refresh(self.biothing_type)
        meta = self.metadata.get_metadata(self.biothing_type)

        if self.args.raw:
            raise Finish(info)

        elif self.args.dev:
            meta["software"] = self.biothings.devinfo.get()

        else:  # remove debug info
            for field in list(meta):
                if field.startswith("_"):
                    meta.pop(field, None)

        if iscoroutinefunction(self.extras):
            meta = await self.extras(meta)
        else:
            meta = self.extras(meta)

        self.finish(dict(sorted(meta.items())))

    def extras(self, _meta):
        """
        Override to add app specific metadata.
        """
        return _meta


class MetadataFieldHandler(BaseQueryHandler):
    """
    GET /metadata/fields
    """

    name = "fields"
    kwargs = dict(BaseQueryHandler.kwargs)
    kwargs["GET"] = {
        "raw": {"type": bool, "default": False},
        "search": {"type": str, "default": None},
        "prefix": {"type": str, "default": None},
    }

    async def get(self):
        await self.metadata.refresh(self.biothing_type)
        mapping = self.metadata.get_mappings(self.biothing_type)

        if self.args.raw:
            raise Finish(mapping)

        result = self.pipeline.formatter.transform_mapping(mapping, self.args.prefix, self.args.search)

        self.finish(result)


async def ensure_awaitable(obj):
    if isinstance(obj, CoroutineType):
        return await obj
    return obj


def capture_exceptions(coro):
    async def _method(*args, **kwargs):
        try:
            return await coro(*args, **kwargs)
        except QueryPipelineInterrupt as itr:
            raise Finish(itr.details)
        except QueryPipelineException as exc:
            raise HTTPError(exc.code, None, exc.details, reason=exc.summary)

    return _method


class BiothingHandler(BaseQueryHandler):
    r"""
    Biothings Annotation Endpoint

    URL pattern examples:

        /{pre}/{ver}/{typ}/?
        /{pre}/{ver}/{typ}/([^\/]+)/?

        queries a term against a pre-determined field that
        represents the id of a document, like _id and dbsnp.rsid

        GET -> {...} or [{...}, ...]
        POST -> [{...}, ...]
    """
    name = "annotation"

    @capture_exceptions
    async def post(self, *args, **kwargs):
        self.event["value"] = len(self.args["id"])

        result = await ensure_awaitable(self.pipeline.fetch(**self.args))
        self.finish(result)

    @capture_exceptions
    async def get(self, *args, **kwargs):
        self.event["value"] = 1

        result = await ensure_awaitable(self.pipeline.fetch(**self.args))
        self.finish(result)


class QueryHandler(BaseQueryHandler):
    """
    Biothings Query Endpoint

    URL pattern examples:

        /{pre}/{ver}/{typ}/query/?
        /{pre}/{ver}//query/?

        GET -> {...}
        POST -> [{...}, ...]
    """

    name = "query"

    @capture_exceptions
    async def post(self, *args, **kwargs):
        self.event["value"] = len(self.args["q"])

        result = await ensure_awaitable(self.pipeline.search(**self.args))
        self.finish(result)

    @capture_exceptions
    async def get(self, *args, **kwargs):
        self.event["value"] = 1
        if self.args.get("fetch_all"):
            self.event["label"] = "fetch_all"

        if self.args.get("fetch_all") or self.args.get("scroll_id") or self.args.get("q") == "__any__":
            self.clear_header("Cache-Control")

        response = await ensure_awaitable(self.pipeline.search(**self.args))
        self.finish(response)



from biothings.hub.dataindex.snapshot_repo import Repository


class Snapshot:
    def __init__(self, client, repository, snapshot):
        # Corresponds to HTTP operations on
        # /_snapshot/<repository>/<snapshot>

        self.client = client
        self.repository = Repository(client, repository)
        self.name = snapshot

    def exists(self):
        if self.repository.exists():
            return bool(
                self.client.snapshot.get(
                    self.repository.name,
                    self.name,
                    ignore_unavailable=True,
                )["snapshots"]
            )
        return False

    def create(self, indices):
        self.client.snapshot.create(
            self.repository.name,
            self.name,
            {
                "indices": indices,
                "include_global_state": False,
            },
        )

    def state(self):
        if self.repository.exists():
            snapshots = self.client.snapshot.get(
                self.repository.name,
                self.name,
                ignore_unavailable=True,
            )["snapshots"]

            if snapshots:  # [{...}]
                return snapshots[0]["state"]
            return "MISSING"

        return "N/A"

    def delete(self):
        self.client.snapshot.delete(self.repository.name, self.name)

    def __str__(self):
        return f"<Snapshot {self.state()} name='{self.name}' repository={self.repository}>"


def test_01():
    from elasticsearch import Elasticsearch

    client = Elasticsearch()
    snapshot = Snapshot(client, "mynews", "mynews_202012280220_vsdevjdk")
    print(snapshot)


def test_02():
    from elasticsearch import Elasticsearch

    client = Elasticsearch()
    snapshot = Snapshot(client, "mynews", "____________________________")
    print(snapshot)


if __name__ == "__main__":
    test_02()



import logging
import mimetypes
import os
import tempfile
import warnings
from urllib.parse import quote

import boto3
import botocore.exceptions

try:
    from biothings import config
except ImportError:
    # assuming key, secret and bucket will be passed
    # to all functions
    pass


def key_exists(bucket, s3key, aws_key=None, aws_secret=None):
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    try:
        client.head_object(Bucket=bucket, Key=s3key)
        return True
    except Exception as e:
        if e.response["Error"]["Code"] == "404":
            return False
        else:
            raise


def send_s3_file(
    localfile,
    s3key,
    overwrite=False,
    permissions=None,
    metadata=None,
    content=None,
    content_type=None,
    aws_key=None,
    aws_secret=None,
    s3_bucket=None,
    redirect=None,
):
    """save a localfile to s3 bucket with the given key.
    bucket is set via S3_BUCKET
    it also save localfile's lastmodified time in s3 file's metadata

    Args:
        redirect (str): if not None, set the redirect property
            of the object so it produces a 301 when accessed
    """
    metadata = metadata or {}
    try:
        aws_key = aws_key or config.AWS_SECRET
        aws_secret = aws_secret or config.AWS_SECRET
        s3_bucket = s3_bucket or config.S3_BUCKET
    except AttributeError:
        logging.info("Skip sending file to S3, missing information in config file: AWS_KEY, AWS_SECRET or S3_BUCKET")
        return
    s3 = boto3.resource("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    target_object = s3.Object(s3_bucket, s3key)
    if not overwrite:
        if key_exists(bucket=s3_bucket, s3key=s3key, aws_key=aws_key, aws_secret=aws_secret):
            # NOTE: change to assert/AssertionError if something relies on
            #  the assert statement
            raise FileExistsError('s3key "{}" already exists.'.format(s3key))
    # assuming metadata is a Mapping type
    put_request = {"Metadata": metadata}
    if redirect:
        put_request["WebsiteRedirectLocation"] = redirect
    if content_type:
        put_request["ContentType"] = content_type
    if content is not None:
        put_request["Body"] = content
    else:
        assert os.path.exists(localfile), 'localfile "{}" does not exist.'.format(localfile)
        lastmodified = os.stat(localfile)[-2]
        put_request["Body"] = open(localfile, "rb")
        put_request["Metadata"]["lastmodified"] = str(lastmodified)
    target_object.put(**put_request)

    if permissions:
        target_object.Acl().put(ACL=permissions)


def send_s3_big_file(
    localfile,
    s3key,
    overwrite=False,
    acl=None,
    aws_key=None,
    aws_secret=None,
    s3_bucket=None,
    storage_class=None,
):
    """
    Multiparts upload for file bigger than 5GiB
    """
    # TODO: maybe merge with send_s3_file() based in file size ? It would need boto3 migration
    try:
        aws_key = aws_key or config.AWS_SECRET
        aws_secret = aws_secret or config.AWS_SECRET
        s3_bucket = s3_bucket or config.S3_BUCKET
    except AttributeError:
        logging.info("Skip sending file to S3, missing information in config file: AWS_KEY, AWS_SECRET or S3_BUCKET")
        return
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    if not overwrite and key_exists(s3_bucket, s3key, aws_key, aws_secret):
        raise Exception("Key '%s' already exist" % s3key)
    tfr_config = boto3.s3.transfer.TransferConfig(
        multipart_threshold=1024 * 25,
        max_concurrency=10,
        multipart_chunksize=1024 * 25,
        use_threads=True,
    )
    extra = {
        "ACL": acl or "private",
        "ContentType": mimetypes.MimeTypes().guess_type(localfile)[0] or "binary/octet-stream",
        "StorageClass": storage_class or "REDUCED_REDUNDANCY",
    }
    client.upload_file(Filename=localfile, Bucket=s3_bucket, Key=s3key, ExtraArgs=extra, Config=tfr_config)


def get_s3_file(s3key, localfile=None, return_what=False, aws_key=None, aws_secret=None, s3_bucket=None):
    # get_s3_file is planned to be deprecated in 0.11 and removed in 0.13
    warnings.warn(
        DeprecationWarning("get_s3_file is deprecated, use download_s3_file or get_s3_file_contents instead"),
        stacklevel=2,
    )

    if return_what == "content":
        return get_s3_file_contents(s3key, aws_key, aws_secret, s3_bucket)
    elif return_what == "key":
        warnings.warn(
            DeprecationWarning("get_s3_file: return_what=key is deprecated, use other ways instead"), stacklevel=2
        )
        try:
            # pylint:disable=import-outside-toplevel
            # this is so that only those who need return_what="key"
            # will depend on boto
            from boto import connect_s3

            # pylint:enable=import-outside-toplevel
            s3 = connect_s3(aws_key, aws_secret)
            bucket = s3.get_bucket(s3_bucket)
            k = bucket.get_key(s3key)
            return k
        except ImportError:
            raise RuntimeError("get_s3_file: return_what=key needs package boto to be installed")
    else:
        download_s3_file(s3key, localfile, aws_key, aws_secret, s3_bucket, overwrite=True)


def _populate_s3_info(aws_key, aws_secret, s3_bucket):
    aws_key = aws_key or getattr(config, "AWS_SECRET", None)
    aws_secret = aws_secret or getattr(config, "AWS_SECRET", None)
    s3_bucket = s3_bucket or getattr(config, "S3_BUCKET", None)
    return aws_key, aws_secret, s3_bucket


def _get_s3_object(aws_key, aws_secret, s3_bucket, s3key):
    aws_key, aws_secret, s3_bucket = _populate_s3_info(aws_key, aws_secret, s3_bucket)
    if not key_exists(s3_bucket, s3key, aws_key, aws_secret):
        raise FileNotFoundError(s3key)
    s3 = boto3.resource("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    target_object = s3.Object(s3_bucket, s3key)
    return target_object


# pylint:disable=too-many-arguments
# at the moment we do not intend to merge parameters (to sth. like S3Config)
def download_s3_file(s3key, localfile=None, aws_key=None, aws_secret=None, s3_bucket=None, overwrite=False):
    localfile = localfile or os.path.basename(s3key)
    if not overwrite and os.path.exists(localfile):
        raise FileExistsError(f"download_s3_file: {localfile} already exists and not overwriting")
    target_object = _get_s3_object(aws_key, aws_secret, s3_bucket, s3key)
    with tempfile.NamedTemporaryFile("xb", delete=False) as tmp:
        body = target_object.get()["Body"]
        for chunk in body.iter_chunks():
            tmp.write(chunk)
        if overwrite:
            os.replace(tmp.name, localfile)
        else:
            os.rename(tmp.name, localfile)


# pylint:enable=too-many-arguments
def get_s3_file_contents(s3key, aws_key=None, aws_secret=None, s3_bucket=None) -> bytes:
    target_object = _get_s3_object(aws_key, aws_secret, s3_bucket, s3key)
    return target_object.get()["Body"].read()


def get_s3_folder(s3folder, basedir=None, aws_key=None, aws_secret=None, s3_bucket=None):
    aws_key = aws_key or config.AWS_SECRET
    aws_secret = aws_secret or config.AWS_SECRET
    s3_bucket = s3_bucket or config.S3_BUCKET
    s3 = boto3.resource("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    bucket = s3.Bucket(s3_bucket)
    cwd = os.getcwd()
    try:
        if basedir:
            os.chdir(basedir)
        if not os.path.exists(s3folder):
            os.makedirs(s3folder)
        for k in bucket.objects.filter(Prefix=s3folder):
            download_s3_file(
                k.key, localfile=k.key, aws_key=aws_key, aws_secret=aws_secret, s3_bucket=s3_bucket, overwrite=True
            )
    finally:
        os.chdir(cwd)


def send_s3_folder(
    folder,
    s3basedir=None,
    acl=None,
    overwrite=False,
    aws_key=None,
    aws_secret=None,
    s3_bucket=None,
):
    aws_key = aws_key or config.AWS_SECRET
    aws_secret = aws_secret or config.AWS_SECRET
    s3_bucket = s3_bucket or config.S3_BUCKET
    s3 = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    s3.head_bucket(Bucket=s3_bucket)  # will raise when not 200
    cwd = os.getcwd()
    if not s3basedir:
        s3basedir = os.path.basename(cwd)
    for localf in [f for f in os.listdir(folder) if not f.startswith(".")]:
        fullpath = os.path.join(folder, localf)
        if os.path.isdir(fullpath):
            send_s3_folder(
                fullpath,
                os.path.join(s3basedir, localf),
                overwrite=overwrite,
                acl=acl,
                aws_key=aws_key,
                aws_secret=aws_secret,
                s3_bucket=s3_bucket,
            )
        else:
            send_s3_big_file(
                fullpath,
                os.path.join(s3basedir, localf),
                overwrite=overwrite,
                acl=acl,
                aws_key=aws_key,
                aws_secret=aws_secret,
                s3_bucket=s3_bucket,
            )


def get_s3_url(s3key, aws_key=None, aws_secret=None, s3_bucket=None):
    if key_exists(s3_bucket, s3key, aws_key, aws_secret):
        return f"https://{s3_bucket}.s3.amazonaws.com/{quote(s3key)}"
    return None


def get_s3_static_website_url(s3key, aws_key=None, aws_secret=None, s3_bucket=None):
    aws_key, aws_secret, s3_bucket = _populate_s3_info(aws_key, aws_secret, s3_bucket)
    s3 = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    location_resp = s3.get_bucket_location(Bucket=s3_bucket)
    region = location_resp.get("LocationConstraint", "us-east-1")
    return f"http://{s3_bucket}.s3-website.{region}.amazonaws.com/{quote(s3key)}"


def create_bucket(name, region=None, aws_key=None, aws_secret=None, acl=None, ignore_already_exists=False):
    """Create a S3 bucket "name" in optional "region". If aws_key and aws_secret
    are set, S3 client will these, otherwise it'll use default system-wide setting.
    "acl" defines permissions on the bucket: "private" (default), "public-read",
    "public-read-write" and "authenticated-read"
    """
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    acl = acl or "private"
    kwargs = {"ACL": acl, "Bucket": name}
    if region:
        kwargs["CreateBucketConfiguration"] = {"LocationConstraint": region}
    try:
        client.create_bucket(**kwargs)
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "BucketAlreadyOwnedByYou" and not ignore_already_exists:
            raise


def set_static_website(name, aws_key=None, aws_secret=None, index="index.html", error="error.html"):
    client = boto3.client("s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret)
    conf = {"IndexDocument": {"Suffix": index}, "ErrorDocument": {"Key": error}}
    client.put_bucket_website(Bucket=name, WebsiteConfiguration=conf)
    location = client.get_bucket_location(Bucket=name)
    region = location["LocationConstraint"]
    # generate website URL
    return "http://%(name)s.s3-website-%(region)s.amazonaws.com" % {"name": name, "region": region}



"""
We can define jmespath custom functions here

See: https://jmespath.org/tutorial.html#custom-functions

from biothings.utils.jmespath import options as jmp_options

jmespath.search("unique(`foo`)", {}, options=jmp_options)
or
jmespath.compile("unique(`foo`)").search({}, options=jmp_options)

"""
import jmespath
from jmespath import functions


class CustomFunctions(functions.Functions):
    """Create a subclass of functions.Functions.
    The function.Functions base class has logic
    that introspects all of its methods and automatically
    registers your custom functions in its function table.
    """

    @functions.signature({"types": ["array"]})
    def _func_unique(self, arr):
        """return a list of unique values in an array"""
        return sorted(set(arr))

    @functions.signature({"types": ["array"]})
    def _func_unique_count(self, arr):
        """return the number of unique values in an array"""
        return len(set(arr))


# pass this jmespath_options to search to use custom functions
options = jmespath.Options(custom_functions=CustomFunctions())



import datetime
from collections import OrderedDict, UserDict, UserList, UserString
from typing import Any, Union
from urllib.parse import parse_qs, unquote_plus, urlencode, urlparse, urlunparse

import orjson
import yaml


def load_json(json_str: Union[bytes, str]) -> Any:
    """Load a JSON string or bytes using orjson"""
    return orjson.loads(json_str)


def to_json_0(data):
    """deprecated"""
    import json

    from biothings.utils.common import BiothingsJSONEncoder

    return json.dumps(data, cls=BiothingsJSONEncoder)


def orjson_default(o):
    """The default function passed to orjson to serialize non-serializable objects"""
    if isinstance(o, (UserDict, UserList)):
        return o.data  # o.data is the actual dictionary of list to store the data
    raise TypeError(f"Type {type(o)} not serializable")


def to_json(data, indent=False, sort_keys=False):
    # default option:
    #    OPT_NON_STR_KEYS: non string dictionary key, e.g. integer
    #    OPT_NAIVE_UTC: use UTC as the timezone when it's missing
    option = orjson.OPT_NON_STR_KEYS | orjson.OPT_NAIVE_UTC
    if indent:
        option |= orjson.OPT_INDENT_2
    if sort_keys:
        option |= orjson.OPT_SORT_KEYS
    return orjson.dumps(data, default=orjson_default, option=option).decode()


def to_json_file(data, fobj, indent=False, sort_keys=False):
    json_str = to_json(data, indent=indent, sort_keys=sort_keys)
    fobj.write(json_str)


# define aliases close to json.loads and json.dumps for convenience
json_loads = load_json
json_dumps = to_json


def to_yaml(data, stream=None, Dumper=yaml.SafeDumper, default_flow_style=False):
    # Author: Cyrus Afrasiabi

    class OrderedDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items())

    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    return yaml.dump(data, stream, OrderedDumper, default_flow_style=default_flow_style)


def to_msgpack(data):
    import msgpack  # import here so that we don't have to install msgpack if we don't use it

    return msgpack.packb(data, use_bin_type=True, default=_msgpack_encode_datetime)


def _msgpack_encode_datetime(obj):
    if isinstance(obj, datetime.datetime):
        return {"__datetime__": True, "as_str": obj.strftime("%Y%m%dT%H:%M:%S.%f")}
    return obj


class URL(UserString):
    def remove(self, param="format"):
        urlparsed = urlparse(str(self))
        qs = parse_qs(urlparsed.query)
        qs.pop(param, None)
        qs = urlencode(qs, True)
        urlparsed = urlparsed._replace(query=qs)
        url = urlunparse(urlparsed)
        return unquote_plus(url)



from typing import Optional

import typer
from typing_extensions import Annotated

from biothings.cli import utils

logger = utils.get_logger("dataplugin")


short_help = (
    "[green]Test an individual data plugin locally and make simple queries to inspect your parsed data objects.[/green]"
)
long_help = (
    short_help
    + "\n\n[magenta]   :sparkles: Go to your existing data plugin folder.[/magenta]"
    + "\n[magenta]   :sparkles: Dumping, uploading and inspecting your data plugin.[/magenta]"
    + "\n[magenta]   :sparkles: Serving your data as a web service for making simple queries[/magenta]"
    + "\n\n[green]   :point_right: Always run this command inside of your data plugin folder.[/green]"
    + "\n[green]   :point_right: Default traceback errors are kept minimal, but you can set [bold]BTCLI_RICH_TRACEBACK=1[/bold][/green]"
    + "\n[green]      ENV variable to enable full and pretty-formatted tracebacks, [/green]"
    + "\n[green]      or set [bold]BTCLI_DEBUG=1[/bold] to enable even more debug logs for debugging purpose.[/green]"
    + "\n[green]   :point_right: You can include a config.py at the working directly to override the default biothings.config settings.[/green]"
    + "\n   :rocket::boom::sparkling_heart:"
)

app = typer.Typer(
    help=long_help,
    short_help=short_help,
    no_args_is_help=True,
    rich_markup_mode="rich",
)


@app.command(
    name="create",
    help="Create a new data plugin from the tempplate",
)
def create_data_plugin(
    name: Annotated[
        str,
        typer.Option("--name", "-n", help="Provide a data plugin name", prompt="What's your data plugin name?"),
    ] = "",
    multi_uploaders: Annotated[
        Optional[bool],
        typer.Option("--multi-uploaders", help="If provided, the data plugin includes multiple uploaders"),
    ] = False,
    # parallelizer: bool = typer.Option(False, "--parallelizer", help="Using parallelizer or not? Default: No"),
    parallelizer: Annotated[
        Optional[bool],
        typer.Option("--parallelizer", help="If provided, the data plugin's upload step will run in parallel"),
    ] = False,
):
    """*create* command for creating a new data plugin from the template"""
    utils.do_create(name, multi_uploaders, parallelizer, logger=logger)


@app.command(
    name="dump",
    help="Download source data files to local",
)
def dump_data():
    """*dump* command for downloading source data files to local"""
    utils.do_dump(plugin_name=None, logger=logger)


@app.command(
    name="upload",
    help="Convert downloaded data from dump step into JSON documents and upload the to the source database",
)
def upload_source(
    batch_limit: Annotated[
        Optional[int],
        typer.Option(
            "--batch-limit",
            help="The maximum number of batches that should be uploaded. Batch size is 1000 docs",
        ),
    ] = None,
):
    """*upload* command for converting downloaded data from dump step into JSON documents and upload the to the source database.
    A local sqlite database used to store the uploaded data"""
    utils.do_upload(plugin_name=None, logger=logger)


@app.command(
    "dump_and_upload",
    help="Download data source to local folder then convert to Json document and upload to the source database",
)
def dump_and_upload(
    # multi_uploaders: bool = typer.Option(
    #     False, "--multi-uploaders", help="Add this option if you want to create multiple uploaders"
    # ),
    # parallelizer: bool = typer.Option(
    #     False, "--parallelizer", help="Using parallelizer or not? Default: No"
    # ),
):
    """*dump_and_upload* command for downloading source data files to local, then converting them into JSON documents and uploading them to the source database.
    Two steps in one command."""
    utils.do_dump_and_upload(plugin_name=None, logger=logger)


@app.command(
    name="list",
    help="Listing dumped files or uploaded sources",
)
def listing(
    dump: Annotated[Optional[bool], typer.Option("--dump", help="Listing dumped files")] = False,
    upload: Annotated[Optional[bool], typer.Option("--upload", help="Listing uploaded sources")] = False,
    hubdb: Annotated[Optional[bool], typer.Option("--hubdb", help="Listing internal hubdb content")] = False,
):
    """*list* command for listing dumped files and/or uploaded sources"""
    utils.do_list(plugin_name=None, dump=dump, upload=upload, hubdb=hubdb, logger=logger)


@app.command(
    name="inspect",
    help="Giving detailed information about the structure of documents coming from the parser",
)
def inspect_source(
    sub_source_name: Annotated[
        Optional[str], typer.Option("--sub-source-name", "-s", help="Your sub source name")
    ] = "",
    mode: Annotated[
        Optional[str],
        typer.Option(
            "--mode",
            "-m",
            help="""
            The inspect mode or list of modes (comma separated), e.g. "type,mapping".\n
            Possible values are:\n
            - "type": explore documents and report strict data structure\n
            - "mapping": same as type but also perform test on data so guess best mapping\n
               (eg. check if a string is splitable, etc...). Implies merge=True\n
            - "stats": explore documents and compute basic stats (count,min,max,sum)\n
            """,
        ),
    ] = "type,stats",
    limit: Annotated[
        Optional[int],
        typer.Option(
            "--limit",
            "-l",
            help="""
            can limit the inspection to the x first docs (None = no limit, inspects all)
            """,
        ),
    ] = None,
    # merge: Annotated[
    #     Optional[bool],
    #     typer.Option(
    #         "--merge",
    #         "-m",
    #         help="""Merge scalar into list when both exist (eg. {"val":..} and [{"val":...}])""",
    #     ),
    # ] = False,
    output: Annotated[
        Optional[str],
        typer.Option(
            "--output",
            "-o",
            help="The local JSON file path for storing mapping info if you run with mode 'mapping' (absolute path or relative path)",
        ),
    ] = None,
):
    """*inspect* command for giving detailed information about the structure of documents coming from the parser after the upload step"""

    utils.do_inspect(
        plugin_name=None,
        sub_source_name=sub_source_name,
        mode=mode,
        limit=limit,
        merge=False,
        output=output,
        logger=logger,
    )


@app.command(name="serve")
def serve(
    host: Annotated[
        Optional[str],
        typer.Option(
            "--host",
            help="The host name to run the test API server",
        ),
    ] = "localhost",
    port: Annotated[
        Optional[int],
        typer.Option(
            "--port",
            "-p",
            help="The port number to tun the test API server",
        ),
    ] = 9999,
):
    """
    *serve* command runs a simple API server for serving documents from the source database.

    For example, after run 'dump_and_upload', we have a source_name = "test" with a document structure
    like this:

    doc = {"_id": "123", "key": {"a":{"b": "1"},"x":[{"y": "3", "z": "4"}, "5"]}}.

    An API server will run at http://host:port/<your source name>/, like http://localhost:9999/test/:

        - You can see all available sources on the index page: http://localhost:9999/
        - You can list all docs: http://localhost:9999/test/ (default is to return the first 10 docs)
        - You can paginate doc list: http://localhost:9999/test/?start=10&limit=10
        - You can retrieve a doc by id: http://localhost:9999/test/123
        - You can filter out docs with one or multiple fielded terms:
            - http://localhost:9999/test/?q=key.a.b:1 (query by any field with dot notation like key.a.b=1)
            - http://localhost:9999/test/?q=key.a.b:1%20AND%20key.x.y=3 (find all docs that match two fields)
            - http://localhost:9999/test/?q=key.x.z:4*  (field value can contain wildcard * or ?)
            - http://localhost:9999/test/?q=key.x:5&start=10&limit=10 (pagination also works)
    """
    utils.do_serve(plugin_name=None, host=host, port=port, logger=logger)


@app.command(
    name="clean",
    help="Delete all dumped files and drop uploaded sources tables",
    no_args_is_help=True,
)
def clean_data(
    dump: Annotated[Optional[bool], typer.Option("--dump", help="Delete all dumped files")] = False,
    upload: Annotated[Optional[bool], typer.Option("--upload", help="Drop uploaded sources tables")] = False,
    clean_all: Annotated[
        Optional[bool],
        typer.Option(
            "--all",
            help="Delete all dumped files and drop uploaded sources tables",
        ),
    ] = False,
):
    """*clean* command for deleting all dumped files and/or drop uploaded sources tables"""
    utils.do_clean(plugin_name=None, dump=dump, upload=upload, clean_all=clean_all, logger=logger)



import pydoc
from functools import partial
from inspect import signature

method_directive = ".. py:method::"
data_directive = ".. py:data::"
hub_commands_title = "biothings.hub.commands\n==============="
hub_commands_short_description = (
    "This document will show you all available commands "
    "that can be used when you access the Hub shell, and their usages."
)
template = """{directive} {name}{signature}

{docstring}
"""


def generate_command_documentations(filepath, commands):
    from biothings.hub import HubCommands

    assert isinstance(commands, (dict, HubCommands)), "commands must be a HubCommands instance, or dict"

    command_docs = []

    for command_name, command_data in sorted(commands.items()):
        func = None
        sign = None
        docstring = ""
        directive = data_directive

        if callable(command_data):
            func = command_data
        elif isinstance(command_data, dict):
            func = command_data.get("command")

        if func:
            if callable(func):
                directive = method_directive
                if isinstance(func, partial):
                    func = func.func
                sign = signature(func)
                docstring = pydoc.render_doc(func, title="%s", renderer=pydoc.plaintext)
            else:
                docstring = "This is a instance of type: {}".format(type(func))

        command_docs.append(
            template.format(
                directive=directive,
                name=command_name,
                signature=str(sign) if sign else "",
                docstring=docstring,
            )
        )

    command_docs = "\n\n".join(command_docs)
    doc = f"{hub_commands_title}\n\n{hub_commands_short_description}\n\n{command_docs}"

    with open(filepath, mode="w") as f:
        f.write(doc)



"""
Utils to compare two list of gene documents, no require to setup Biothings Hub.
"""
import json
import time

from .common import filter_dict, timesofar


def diff_doc(doc_1, doc_2, exclude_attrs=None):
    exclude_attrs = exclude_attrs or ["_timestamp"]

    diff_d = {"update": {}, "delete": [], "add": {}}
    if exclude_attrs:
        doc_1 = filter_dict(doc_1, exclude_attrs)
        doc_2 = filter_dict(doc_2, exclude_attrs)
    for attr in set(doc_1) | set(doc_2):
        if exclude_attrs and attr in exclude_attrs:
            continue
        if attr in doc_1 and attr in doc_2:
            _v1 = doc_1[attr]
            _v2 = doc_2[attr]
            if _v1 != _v2:
                diff_d["update"][attr] = _v2
        elif attr in doc_1 and attr not in doc_2:
            diff_d["delete"].append(attr)
        else:
            diff_d["add"][attr] = doc_2[attr]
    if diff_d["update"] or diff_d["delete"] or diff_d["add"]:
        return diff_d


def full_diff_doc(doc_1, doc_2, exclude_attrs=None):
    exclude_attrs = exclude_attrs or ["_timestamp"]

    diff_d = {"update": {}, "delete": [], "add": {}}
    if exclude_attrs:
        doc_1 = filter_dict(doc_1, exclude_attrs)
        doc_2 = filter_dict(doc_2, exclude_attrs)
    for attr in set(doc_1) | set(doc_2):
        if exclude_attrs and attr in exclude_attrs:
            continue
        if attr in doc_1 and attr in doc_2:
            _v1 = doc_1[attr]
            _v2 = doc_2[attr]
            difffound = False
            if isinstance(_v1, dict) and isinstance(_v2, dict):
                if full_diff_doc(_v1, _v2, exclude_attrs):
                    difffound = True
            elif isinstance(_v1, list) and isinstance(_v2, list):
                # there can be unhashable/unordered dict in these lists
                for i in _v1:
                    if i not in _v2:
                        difffound = True
                        break
                # check the other way
                if not difffound:
                    for i in _v2:
                        if i not in _v1:
                            difffound = True
                            break
            elif _v1 != _v2:
                difffound = True

            if difffound:
                diff_d["update"][attr] = _v2

        elif attr in doc_1 and attr not in doc_2:
            diff_d["delete"].append(attr)
        else:
            diff_d["add"][attr] = doc_2[attr]
    if diff_d["update"] or diff_d["delete"] or diff_d["add"]:
        return diff_d


def two_docs_iterator(docs1, docs2, id_list, step=10000, verbose=False):
    t0 = time.time()
    n = len(id_list)
    for i in range(0, n, step):
        t1 = time.time()
        if verbose:
            print("Processing %d-%d documents..." % (i + 1, min(i + step, n)), end="")
        _ids = id_list[i : i + step]
        iter1 = sorted([doc for doc in docs1 if doc["_id"] in _ids], key=lambda a: a["_id"])
        iter2 = sorted([doc for doc in docs2 if doc["_id"] in _ids], key=lambda a: a["_id"])
        for doc1, doc2 in zip(iter1, iter2):
            yield doc1, doc2
        if verbose:
            print("Done.[%.1f%%,%s]" % (i * 100.0 / n, timesofar(t1)))
    if verbose:
        print("=" * 20)
        print("Finished.[total time: %s]" % timesofar(t0))


def diff_docs(docs1, docs2, ids, fastdiff=False, diff_func=full_diff_doc):
    """if fastdiff is True, only compare the whole doc,
    do not traverse into each attributes.
    """
    _updates = []
    for doc1, doc2 in two_docs_iterator(docs1, docs2, ids):
        assert doc1["_id"] == doc2["_id"], repr((ids, len(ids)))
        if fastdiff:
            if doc1 != doc2:
                _updates.append({"_id": doc1["_id"]})
        else:
            _diff = diff_func(doc1, doc2)
            if _diff:
                _diff["_id"] = doc1["_id"]
                _updates.append(_diff)
    return _updates


def normalize_document(docs):
    """
    If docs which is generated by MongoExport, or parsed by DataPlugin, is already in right structure.
    Only need to restructure ES Index data.

    Example output:
    [
        {
            "_id": "PA123",
            "source_name": [
                ...
            ],
        },
    ]

    """
    if not isinstance(docs, (list, tuple)):
        raise Exception("Invalid docs. Expect a list of dict")

    if len(docs) == 0:
        return []

    first_item = docs[0]
    assert "_id" in first_item, "Data struct is invalid. Missing _id field"

    result = docs
    # If data is an json exported from ES Index, then restructure it to be like datasource.
    if "_index" in first_item:
        result = [{"_id": doc["_id"], **doc["_source"]} for doc in docs]

    return result


def get_id_set(docs):
    ids = [doc["_id"] for doc in docs]
    return set(ids)


def diff_collections(docs1, docs2):
    """
    data1, data2 are one of exported data from mongo's collection, or ES index

    To export data from mongo, should use mongoexport command with jsonArray and type=json flags,
    in order to generate valid json file
    """

    _docs1 = normalize_document(docs1)
    _docs2 = normalize_document(docs2)

    id_s1 = get_id_set(_docs1)
    id_s2 = get_id_set(_docs2)
    print("Size of collection 1:\t", len(id_s1))
    print("Size of collection 2:\t", len(id_s2))

    id_in_1 = id_s1 - id_s2
    id_in_2 = id_s2 - id_s1
    id_common = id_s1 & id_s2
    print("# of docs found only in collection 1:\t", len(id_in_1))
    print("# of docs found only in collection 2:\t", len(id_in_2))
    print("# of docs found in both collections:\t", len(id_common))

    print("Comparing matching docs...")
    _updates = []
    if len(id_common) > 0:
        _updates = diff_docs(_docs1, _docs2, list(id_common))
        print("Done. [{} docs changed]".format(len(_updates)))

    _deletes = []
    if len(id_in_1) > 0:
        _deletes = sorted(id_in_1)

    _adds = []
    if len(id_in_2) > 0:
        _adds = sorted(id_in_2)

    changes = {"update": _updates, "delete": _deletes, "add": _adds}
    return changes


def diff_json_files(fpath1, fpath2):
    print("Loading data from json files")
    with open(fpath1) as f:
        docs1 = json.load(f)

    with open(fpath2) as f:
        docs2 = json.load(f)

    print("Do compare 2 docs")
    return diff_collections(docs1, docs2)



from elasticsearch import AsyncElasticsearch, Elasticsearch


class _HCResult:
    MIN_RES_FLD = "status"

    def __init__(self, verbose=False):
        self._verbose = verbose
        self._content = dict(success=True)

    def update(self, dic):
        if self._verbose:
            self._content.update(dic)
        elif self.MIN_RES_FLD in dic:
            self._content[self.MIN_RES_FLD] = dic[self.MIN_RES_FLD]

    def to_dict(self):
        return dict(self._content)


class DBHealth:
    def __init__(self, client):
        self.client = client

    def check(self, **kwargs):
        # raise Exception in case of an error
        # return db server status details.
        raise NotImplementedError()


class ESHealth(DBHealth):
    # https://www.elastic.co/guide/en/elasticsearch/reference/current/cluster-health.html
    # GET http://localhost:9200/_cluster/health
    # {
    #     "cluster_name": "docker-cluster",
    #     "status": "yellow",
    #     "timed_out": false,
    #     "number_of_nodes": 1,
    #     "number_of_data_nodes": 1,
    #     "active_primary_shards": 5,
    #     "active_shards": 5,
    #     "relocating_shards": 0,
    #     "initializing_shards": 0,
    #     "unassigned_shards": 1,
    #     "delayed_unassigned_shards": 0,
    #     "number_of_pending_tasks": 0,
    #     "number_of_in_flight_fetch": 0,
    #     "task_max_waiting_in_queue_millis": 0,
    #     "active_shards_percent_as_number": 83.33333333333334
    # }

    # Future Work A
    # Add transport level cluster information,
    # like connected nodes, etc

    # Future Work B
    # It is useful to provide two endpoints,
    # one indicating the service status, factoring in more
    # measurement dimensions like cluster health and data integrity,
    # primarily used for internal error reporting and operations,
    # and another one indicating just the web application health,
    # indicating if we need to apply resource provision level
    # failure resolution strategy like instance auto re-launch.

    def __init__(self, client, payload=None):
        self.client = client
        self.payload = payload
        # {
        #   "index": "genedoc_current",
        #   "id": "1017"
        # }

    async def async_check(self, verbose=False):
        assert isinstance(self.client, AsyncElasticsearch)

        # this endpoint might be accessed frequently,
        # keep the default response minimal by default.

        response = _HCResult(verbose)
        response.update(await self.client.cluster.health())

        if verbose:  # request additional information
            response.update(await self.client.info())

        if self.payload:
            document = await self.client.get(**self.payload)
            response.update({"payload": self.payload})
            response.update({"document": document})

        return response.to_dict()

    def check(self):
        assert isinstance(self.client, Elasticsearch)
        return self.client.cluster.health()


class MongoHealth(DBHealth):
    def check(self, **kwargs):
        # typical response: {'ok': 1.0}
        return self.client.command("ping")


class SQLHealth(DBHealth):
    def check(self, **kwargs):
        # https://docs.sqlalchemy.org/en/13/core/connections.html
        # #sqlalchemy.engine.Connection.closed
        return dict(closed=self.client.closed)



import importlib.util
import logging
import os
import types
from collections.abc import Collection
from importlib import import_module
from typing import NamedTuple

from . import default, validators

logger = logging.getLogger(__name__)


def load(config="config"):
    config = load_module(config)
    if config.__name__ == config.__package__:
        attrs = [getattr(config, attr) for attr in dir(config)]
        confs = [attr for attr in attrs if isinstance(attr, types.ModuleType)]
        valis = (
            validators.WebAPIValidator(),
            validators.DBParamValidator(),
            validators.SubmoduleValidator(),
        )
        return ConfigPackage(
            ConfigModule(config),
            [ConfigModule(conf, config, valis) for conf in confs],
        )
    else:  # config is a single file module, not a package
        return ConfigModule(
            config,
            validators=(
                validators.WebAPIValidator(),
                validators.DBParamValidator(),
            ),
        )


def load_module(config, default=None):
    """
    Load a config module.

    config:
        1. a module object
        2. a fully qualified module name
        3. a file path to a module
    """
    if isinstance(config, types.ModuleType):
        return config
    elif isinstance(config, str) and config.endswith(".py"):
        spec = importlib.util.spec_from_file_location("config", config)
        config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(config)
        return config
    elif isinstance(config, str) and config:
        return import_module(config)
    elif not config and default:
        return default
    raise ValueError(str(config))


# Opinion on item access []
# The ConfigModule is intended to present a interface similar to
# types.ModuleType or types.SimpleNamespace, and hold primarily
# known attributes that persisit throughout the application. For
# those reasons, item-access interface is not supported.


class ConfigPackage(NamedTuple):
    root: object
    modules: Collection


class ConfigModule:
    """
    A wrapper for the settings that configure the web API.

    * Environment variables can override settings of the same names.
    * Default values are defined in biothings.web.settings.default.

    """

    def __init__(self, config=None, parent=None, validators=(), **kwargs):
        self._fallback = parent  # config package
        self._primary = config
        self._override = types.SimpleNamespace()

        logger.info("%s", self._primary)  # log file location

        # process keyword setting override
        for key, value in kwargs.items():
            setattr(self._override, key, value)

        # process environment variable override of named settings
        for name in os.environ:
            if hasattr(self, name):
                new_value = None
                if isinstance(getattr(self, name), str):
                    new_value = os.environ[name]
                elif isinstance(getattr(self, name), bool):
                    new_value = os.environ[name].lower() in ("true", "1")
                if new_value is not None:
                    logger.info("$ %s = %s", name, os.environ[name])
                    setattr(self._override, name, new_value)
                else:  # cannot override dict, array, object type...
                    logger.error("Env %s is not suppored.", name)

        for validator in validators:
            validator.validate(self)

    def __getattr__(self, name):
        # transient settings like envs
        if hasattr(self._override, name):
            return getattr(self._override, name)
        # user specified config module
        elif hasattr(self._primary, name):
            return getattr(self._primary, name)
        # shared settings in a config package
        elif hasattr(self._fallback, name):
            return getattr(self._fallback, name)
        # global default settings
        elif hasattr(default, name):
            return getattr(default, name)
        else:  # not provided and no default
            raise AttributeError(name)



"""
    Biothings Test Helpers

    There are two types of test classes that provide utilities to
    three types of test cases, developed in the standalone apps.

    The two types of test classes are:
        BiothingsWebTest, which targets a running web server.
        BiothingsWebAppTest, which targets a web server config file.

    To further illustrate, for any biothings web applications, it
    typically conforms to the following architectures:

    Layer 3: A web server that implements the behaviors defined below.
    Layer 2: A config file that defines how to serve data from ES.
    Layer 1: An Elasticsearch server with data.

    And for the two types of test classes, to explain their differences
    in the context of the layered design described above:
        BiothingsWebTest targets an existing Layer 3 endpoint.
        BiothingsWebAppTest targets layer 2 and runs its own layer 3.
        Note no utility is provided to directly talk to layer 1.

    The above discussed the python structures provided as programming
    utilities, on the other hand, there are three types of use cases,
    or testing objectives:
        L3 Data test, which is aimed to test the data integrity of an API.
            It subclasses BiothingsWebTest and ensures all layers working.
            The data has to reside in elasticsearch already.
        L3 Feature test, which is aimed to test the API implementation.
            It makes sure the settings in config file is reflected.
            These tests work on production data and require constant
            updates to keep the test cases in sync with the actual data.
            These test cases subclass BiothingsWebTest as well and asl
            require existing production data in elasticsearch.
        L2 Feature test, doing basically the same things as above but uses
            a small set of data that it ingests into elasticsearch.
            This is a lightweight test for development and automated testings
            for each new commit. It comes with data it will ingest in ES
            and does not require any existing data setup to run.

    To illustrate the differences in a chart:
    +--------------+---------------------+-----------------+-------------+---------------------------+
    | Objectives   | Class               | Test Target     | ES Has Data | Automated Testing Trigger |
    +--------------+---------------------+-----------------+-------------+---------------------------+
    | L3 Data Test | BiothingsWebTest    | A Running API   | Yes         | Data Release              |
    +--------------+---------------------+-----------------+-------------+---------------------------+
    | L3 Feature T.| BiothingsWebTest    | A Running API   | Yes         | Data Release & New Commit |
    +--------------+---------------------+-----------------+-------------+---------------------------+
    | L2 Feature T.| BiothingsWebAppTest | A config module | No*         | New Commit                |
    +--------------+---------------------+-----------------+-------------+---------------------------+
    * For L2 Feature Test, data is defined in the test cases and will be automatically ingested into
      Elasticsearch at the start of the testing and get deleted after testing finishes. The other
      two types of testing require existing production data on the corresponding ES servers.

    In development, it is certainly possible for a particular test case
    to fall under multiple test types, then the developer can use proper
    inheritance structures to avoid repeating the specific test case.

    In terms of naming conventions, sometimes the L3 tests are grouped
    together and called remote tests, as they mostly target remote servers.
    And the L2 tests are called local tests, as they starts a local server.

    L3 Envs:

    TEST_SCHEME
    TEST_PREFIX
    TEST_HOST
    TEST_CONF

    L2 Envs:

    TEST_KEEPDATA
    < Config Module Override >

"""
import glob
import inspect
import json
import os
import sys
from functools import partial
from typing import Optional, Union

import pytest
import requests
import urllib3
from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase

import biothings
from biothings.utils.common import traverse
from biothings.web.launcher import BiothingsAPI
from biothings.web.settings import configs


class BiothingsWebTestBase:
    # relative path parsing configuration
    scheme = "http"
    prefix = "v1"
    host = ""

    def request(self, path, method="GET", expect=200, **kwargs):
        """
        Use requests library to make an HTTP request.
        Ensure path is translated to an absolute path.
        Conveniently check if status code is as expected.
        """
        url = self.get_url(path)
        res = requests.request(method, url, **kwargs)
        assert res.status_code == expect, res.text

        return res

    def get_url(self, path):
        """
        Try best effort to get a full url to make a request.
        Return an absolute url when class var 'host' is defined.
        If not, return a path relative to the host root.
        """
        # already an absolute path
        if path.lower().startswith(("http://", "https://")):
            return path

        # path standardization
        if not path.startswith("/"):
            if self.prefix:  # append prefix
                path = "/".join((self.prefix, path))
            path = "/" + path

        # host standardization
        if self.host:
            path = f"{self.scheme}://{self.host}{path}"

        return path

    def query(self, method="GET", endpoint="query", hits=True, data=None, json=None, **kwargs):
        """
        Make a query and assert positive hits by default.
        Assert zero hit when hits is set to False.
        """

        if method == "GET":
            res = self.request(endpoint, params=kwargs, data=data, json=json).json()

            assert bool(res.get("hits")) == hits
            return res

        if method == "POST":
            res = self.request(endpoint, method=method, params=kwargs, data=data, json=json).json()

            for item in res:  # list
                if "_id" not in item:
                    _hits = False
                    break
            else:
                _hits = bool(res)
            assert _hits is hits
            return res

        raise ValueError("Invalid Request Method.")

    @staticmethod
    def msgpack_ok(packed_bytes):
        """Load msgpack into a dict"""
        try:
            import msgpack
        except ImportError:
            pytest.skip("Msgpack is not installed.")
        try:
            dic = msgpack.unpackb(packed_bytes)
        except BaseException:  # pylint: disable=bare-except
            # assert False, "Not a valid Msgpack binary."
            raise ValueError("Not a valid Msgpack binary.")
        return dic

    @staticmethod
    def value_in_result(value, result: Union[dict, list], key: str, case_insensitive: bool = False) -> bool:
        """
        Check if value is in result at specific key

        Elasticsearch does not care if a field has one or more values (arrays),
        so you may get a search with multiple values in one field.
        You were expecting a result of type T but now you have a List[T] which
        is bad.
        In testing, usually any one element in the list eq. to the value you're
        looking for, you don't really care which.
        This helper function checks if the value is at a key, regardless
        of the details of nesting, so you can just do this:
            assert self.value_in_result(value, result, 'where.it.should.be')

        Caveats:
        case_insensitive only calls .lower() and does not care about locale/
        unicode/anything

        Args:
            value: value to look for
            result: dict or list of input, most likely from the APIs
            key: dot delimited key notation
            case_insensitive: for str comparisons, invoke .lower() first
        Returns:
            boolean indicating whether the value is found at the key
        Raises:
            TypeError: when case_insensitive set to true on unsupported types
        """
        res_at_key = []
        if case_insensitive:
            try:
                value = value.lower()
            except Exception:
                raise TypeError("failed to invoke method .lower()")
        for k, v in traverse(result, leaf_node=True):
            if k == key:
                if case_insensitive:
                    try:
                        v = v.lower()
                    except Exception:
                        raise TypeError("failed to invoke method .lower()")
                res_at_key.append(v)
        return value in res_at_key


class BiothingsWebTest(BiothingsWebTestBase):
    """ """

    @classmethod
    def setup_class(cls):
        """this is the setup method when pytest run tests from this class"""
        cls.scheme = os.getenv("TEST_SCHEME", cls.scheme)
        cls.prefix = os.getenv("TEST_PREFIX", cls.prefix).strip("/")
        cls.host = os.getenv("TEST_HOST", cls.host).strip("/")
        base_url = f"{cls.scheme}://{cls.host}/{cls.prefix}" if cls.host else f"/{cls.prefix}"
        msg = f"\n\tTest URL: {base_url}"
        msg += f"\n\tBioThings SDK Version: {biothings.__version__}"
        msg += f"\n\tBioThings SDK path: {biothings.__file__}\n"
        # this stderr output will be suppressed by pytest
        # but will be shown when --capture=no or -s is passed,
        # allowing us to see the test url when running tests
        sys.__stderr__.write(msg)


class BiothingsWebAppTest(BiothingsWebTestBase, AsyncHTTPTestCase):
    """
    Starts the tornado application to run tests locally.
    Need a config.py under the test class folder.
    """

    TEST_DATA_DIR_NAME: Optional[str] = None  # set sub-dir name

    @pytest.fixture(scope="class", autouse=True)
    def _setup_elasticsearch(self):
        # Author: Zhongchao Qian
        # https://github.com/biothings/biothings.api/pull/135

        if not self.TEST_DATA_DIR_NAME:
            yield  # do no setup and yield control to pytest
            return

        s = requests.Session()
        s.mount(
            "http://", adapter=requests.adapters.HTTPAdapter(max_retries=urllib3.Retry(total=5, backoff_factor=3.0))
        )  # values seem reasonable
        es_host = "http://" + self.config.ES_HOST

        server_info = s.get(es_host).json()
        version_info = tuple(int(v) for v in server_info["version"]["number"].split("."))
        if version_info[0] < 6 or version_info[0] == 6 and version_info[1] < 8:
            pytest.exit("Tests need to be running on ES6.8+")

        indices = []  # for cleanup later
        data_dir_path = os.path.dirname(inspect.getfile(type(self)))
        data_dir_path = os.path.join(data_dir_path, "test_data")
        data_dir_path = os.path.join(data_dir_path, self.TEST_DATA_DIR_NAME)
        glob_json_pattern = os.path.join(data_dir_path, "*.json")
        # wrap around in try-finally so the index is guaranteed to be
        err_flag = False
        try:
            # TODO No match seems to cause illegible error
            for index_mapping_path in glob.glob(glob_json_pattern):
                index_name = os.path.basename(index_mapping_path)
                index_name = os.path.splitext(index_name)[0]
                indices.append(index_name)
                r = s.head(f"{es_host}/{index_name}")
                if r.status_code != 404:
                    if os.environ.get("TEST_KEEPDATA"):
                        continue
                    raise RuntimeError(f"{index_name} already exists!")
                with open(index_mapping_path, "r") as f:
                    mapping = json.load(f)
                data_path = os.path.join(data_dir_path, index_name + ".ndjson")
                with open(data_path, "r") as f:
                    bulk_data = f.read()
                if version_info[0] == 6:
                    r = s.put(f"{es_host}/{index_name}", json=mapping, params={"include_type_name": "false"})
                elif version_info[0] > 6:
                    r = s.put(f"{es_host}/{index_name}", json=mapping)
                else:
                    raise RuntimeError("This shouldn't have happened")
                r.raise_for_status()
                if version_info[0] < 8:
                    r = s.post(
                        f"{es_host}/{index_name}/_doc/_bulk",
                        data=bulk_data,
                        headers={"Content-type": "application/x-ndjson"},
                    )
                elif version_info[0] >= 8:
                    r = s.post(
                        f"{es_host}/{index_name}/_bulk",
                        data=bulk_data,
                        headers={"Content-type": "application/x-ndjson"},
                    )
                else:
                    raise RuntimeError("This shouldn't have happened")
                r.raise_for_status()
                s.post(f"{es_host}/{index_name}/_refresh")
            yield
        except Exception as e:
            err_msg = str(e)
            err_flag = True
        finally:
            if not os.environ.get("TEST_KEEPDATA"):
                for index_name in indices:
                    s.delete(f"{es_host}/{index_name}")
            if err_flag:
                pytest.exit("Error setting up ES for tests:", err_msg)

    @property
    def config(self):
        if not hasattr(self, "_config"):
            conf = os.getenv("TEST_CONF", "config.py")
            base = os.path.dirname(inspect.getfile(type(self)))
            file = os.path.join(base, conf)
            self._config = configs.load(file)
        return self._config

    # override
    def get_new_ioloop(self):
        return IOLoop.current()

    # override
    def get_app(self):
        prefix = self.config.APP_PREFIX
        version = self.config.APP_VERSION
        self.prefix = f"{prefix}/{version}".strip("/")
        return BiothingsAPI.get_app(self.config)

    # override
    def request(self, path, method="GET", expect=200, **kwargs):
        url = self.get_url(path)

        func = partial(requests.request, method, url, **kwargs)
        res = self.io_loop.run_sync(lambda: self.io_loop.run_in_executor(None, func))

        assert res.status_code == expect, res.text
        return res

    # override
    def get_url(self, path):
        path = BiothingsWebTestBase.get_url(self, path)
        return AsyncHTTPTestCase.get_url(self, path)


# Compatibility
BiothingsTestCase = BiothingsWebAppTest
BiothingsDataTest = BiothingsWebTest



import abc
import asyncio
import copy
import os
from collections import UserDict
from copy import deepcopy
from datetime import datetime
from functools import partial
from typing import NamedTuple, Optional

import elasticsearch
from elasticsearch import AsyncElasticsearch

from biothings import config as btconfig
from biothings.hub import INDEXER_CATEGORY, INDEXMANAGER_CATEGORY
from biothings.hub.databuild.backend import merge_src_build_metadata
from biothings.utils.common import get_class_from_classpath, get_random_string, iter_n, merge, traverse
from biothings.utils.es import ESIndexer
from biothings.utils.hub_db import get_src_build
from biothings.utils.loggers import get_logger
from biothings.utils.manager import BaseManager
from biothings.utils.mongo import DatabaseClient, id_feeder

from .indexer_cleanup import Cleaner
from .indexer_payload import DEFAULT_INDEX_MAPPINGS, DEFAULT_INDEX_SETTINGS, IndexMappings, IndexSettings
from .indexer_registrar import IndexJobStateRegistrar, MainIndexJSR, PostIndexJSR, PreIndexJSR
from .indexer_schedule import Schedule
from .indexer_task import dispatch

# Summary
# -------
# IndexManager: a hub feature, providing top level commands and config environments(env).
# Indexer/ColdHotIndexer: the "index" command, handles jobs, db state and errors.
# .indexer_task.IndexingTask: index a set of ids, running independent of the hub.

# TODO
# Clarify returned result
# Distinguish creates/updates/deletes
# So that hot/cold indexer doc count can be accurate

# TODO
# Multi-layer logging


class IndexerException(Exception):
    ...


class ProcessInfo:
    def __init__(self, indexer, concurrency):
        self.indexer = indexer
        self.concurrency = concurrency

    def get_predicates(self):
        def limit_indexer_concurrency(job_manager):
            def by_indexer_environment(job):
                return all(
                    (
                        job["category"] == INDEXER_CATEGORY,
                        job["source"] == self.indexer.env_name,
                    )
                )

            return len(list(filter(by_indexer_environment, job_manager.jobs.values()))) < self.concurrency

        return [limit_indexer_concurrency]

    def get_pinfo(self, step="", description=""):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "__predicates__": self.get_predicates(),
            "category": INDEXER_CATEGORY,
            "source": self.indexer.env_name,
            "description": description,
            "step": step,
        }
        return pinfo


class _BuildBackend(NamedTuple):  # mongo
    args: dict = {}
    dbs: Optional[str] = None
    col: Optional[str] = None


class _BuildDoc(UserDict):
    """Represent A Build Under "src_build" Collection.

    Example:
    {
        "_id":"mynews_202105261855_5ffxvchx",
        "target_backend": "mongo",
        "target_name": "mynews_202105261855_5ffxvchx", # UNUSED
        "backend_url": "mynews_202105261855_5ffxvchx",
        "build_config": {
            "_id": "mynews",
            "name": "mynews",
            "doc_type": "news",
            ...
            "cold_collection": "mynews_202012280220_vsdevjdk"
        },
        "mapping": {
            "author": {"type": "text" },
            "title": {"type": "text" },
            "description": {"type": "text" },
            ...
        },
        "_meta": {
            "biothing_type": "news",
            "build_version": "202105261855",
            "build_date": "2021-05-26T18:55:00.054622+00:00",
            ...
        },
        ...
    }
    """

    @property
    def build_name(self):
        return self.get("_id")

    @property
    def build_config(self):
        return self.setdefault("build_config", {})

    def enrich_mappings(self, mappings):
        mappings["__hub_doc_type"] = self.build_config.get("doc_type")
        mappings["properties"].update(self.get("mapping", {}))
        mappings["_meta"] = self.get("_meta", {})

    def enrich_settings(self, settings):
        settings["number_of_shards"] = self.build_config.get("num_shards", 1)
        settings["number_of_replicas"] = self.build_config.get("num_replicas", 0)
        # this feature may be removed at any time
        settings.update(self.build_config.get("extra_index_settings", {}))

    def parse_backend(self):
        # Support Sebastian's hub style backend URI
        # #biothings.hub.databuild.backend.create_backend
        backend = self.get("target_backend")
        backend_url = self.get("backend_url")

        # Case 1:
        # As a dummy indexer
        # Used in validate_mapping, ...

        if backend is None:
            return _BuildBackend()

        # Case 2:
        # Most common setup
        # Index a merged collection

        elif backend == "mongo":
            from biothings.hub.databuild import backend

            db = backend.mongo.get_target_db()
            if backend_url in db.list_collection_names():
                return _BuildBackend(
                    dict(zip(("host", "port"), db.client.address)),
                    db.name,
                    backend_url,
                )

        # Case 3:
        # For single source build_config(s)
        # Index the source collection directly

        elif backend == "link":
            from biothings.hub.databuild import backend

            if backend_url[0] == "src":
                db = backend.mongo.get_src_db()
            else:  # backend_url[0] == "target"
                db = backend.mongo.get_target_db()

            if backend_url[1] in db.list_collection_names():
                return _BuildBackend(
                    dict(zip(("host", "port"), db.client.address)),
                    db.name,
                    backend_url[1],
                )

        raise ValueError(backend, backend_url)

    def extract_coldbuild(self):
        cold_target = self.build_config["cold_collection"]
        cold_build_doc = get_src_build().find_one({"_id": cold_target})
        cold_build_doc = _BuildDoc(cold_build_doc)

        cold_build_doc["_id"] = self.build_name  # *
        cold_build_doc["mapping"].update(self["mapping"])  # combine mapping
        merge_src_build_metadata([cold_build_doc, self])  # combine _meta

        # * About State Updates
        # All updates are diverted to the hot collection.
        # Indices & snapshots are only registered there.

        if self.build_config.get("num_shards"):
            cold_build_doc.build_config["num_shards"] = self.build_config["num_shards"]
        if self.build_config.get("num_replicas"):
            cold_build_doc.build_config["num_replicas"] = self.build_config["num_replicas"]
        return cold_build_doc


class Step(abc.ABC):
    name: property(abc.abstractmethod(lambda _: ...))
    state: property(abc.abstractmethod(lambda _: ...))
    method: property(abc.abstractmethod(lambda _: ...))
    catelog = dict()

    @staticmethod
    def order(steps):
        if isinstance(steps, str):
            return (yield from Step.order([steps]))
        for _step in ("pre", "index", "post"):
            if _step in steps:
                yield _step

    def __init__(self, indexer):
        self.indexer = indexer
        self.state = self.state(
            get_src_build(),
            indexer.build_name,
            indexer.es_index_name,
            logfile=indexer.logfile,
        )

    @classmethod
    def __init_subclass__(cls):
        cls.catelog[cls.name] = cls

    @classmethod
    def dispatch(cls, name):
        return cls.catelog[name]

    async def execute(self, *args, **kwargs):
        coro = getattr(self.indexer, self.method)
        coro = coro(*args, **kwargs)
        return await coro

    def __str__(self):
        return f"<Step name='{self.name}' indexer={self.indexer}>"


class PreIndexStep(Step):
    name = "pre"
    state = PreIndexJSR
    method = "pre_index"


class MainIndexStep(Step):
    name = "index"
    state = MainIndexJSR
    method = "do_index"


class PostIndexStep(Step):
    name = "post"
    state = PostIndexJSR
    method = "post_index"


class _IndexerResult(UserDict):
    def __str__(self):
        return f"{type(self).__name__}({str(self.data)})"


class IndexerCumulativeResult(_IndexerResult):
    ...


class IndexerStepResult(_IndexerResult):
    ...


class Indexer:
    """
    MongoDB -> Elasticsearch Indexer.
    """

    def __init__(self, build_doc, indexer_env, index_name):
        # build_doc primarily describes the source.
        # indexer_env primarily describes the destination.

        _build_doc = _BuildDoc(build_doc)
        _build_backend = _build_doc.parse_backend()

        # ----------source----------

        self.mongo_client_args = _build_backend.args
        self.mongo_database_name = _build_backend.dbs
        self.mongo_collection_name = _build_backend.col

        # -----------dest-----------

        # [1] https://elasticsearch-py.readthedocs.io/en/v7.12.0/api.html#elasticsearch.Elasticsearch
        # [2] https://elasticsearch-py.readthedocs.io/en/v7.12.0/helpers.html#elasticsearch.helpers.bulk
        self.es_client_args = indexer_env.get("args", {})  # See [1] for available args
        self.es_blkidx_args = indexer_env.get("bulk", {})  # See [2] for available args
        self.es_index_name = index_name or _build_doc.build_name
        self.es_index_settings = IndexSettings(deepcopy(DEFAULT_INDEX_SETTINGS))
        self.es_index_mappings = IndexMappings(deepcopy(DEFAULT_INDEX_MAPPINGS))

        _build_doc.enrich_settings(self.es_index_settings)
        _build_doc.enrich_mappings(self.es_index_mappings)

        # -----------info-----------

        self.env_name = indexer_env.get("name")
        self.conf_name = _build_doc.build_config.get("name")
        self.build_name = _build_doc.build_name

        self.setup_log()
        self.pinfo = ProcessInfo(self, indexer_env.get("concurrency", 10))

    def setup_log(self):
        log_folder = os.path.join(btconfig.LOG_FOLDER, "build", self.build_name or "", "index") if btconfig.LOG_FOLDER else None
        log_name = f"index_{self.es_index_name}"
        self.logger, self.logfile = get_logger(log_name, log_folder=log_folder, force=True)

    def __str__(self):
        showx = self.mongo_collection_name != self.es_index_name
        lines = [
            f"<{type(self).__name__}",
            f" source='{self.mongo_collection_name}'" if showx else "",
            f" dest='{self.es_index_name}'>",
        ]
        return "".join(lines)

    # --------------
    #  Entry Point
    # --------------
    async def index(self, job_manager, **kwargs):
        """
        Build an Elasticsearch index (self.es_index_name)
        with data from MongoDB collection (self.mongo_collection_name).

        "ids" can be passed to selectively index documents.

        "mode" can have the following values:
            - 'purge': will delete an index if it exists.
            - 'resume': will use an existing index and add missing documents.
            - 'merge': will merge data to an existing index.
            - 'index' (default): will create a new index.
        """

        steps = kwargs.pop("steps", ("pre", "index", "post"))
        batch_size = kwargs.setdefault("batch_size", 10000)
        # mode = kwargs.setdefault("mode", "index")
        kwargs.setdefault("mode", "index")
        ids = kwargs.setdefault("ids", None)

        assert job_manager
        assert all(isinstance(_id, str) for _id in ids) if ids else True
        assert 500 <= batch_size <= 10000, '"batch_size" out-of-range'

        # the batch size here controls only the task partitioning
        # it does not affect how the elasticsearch python client
        # makes batch requests. a number larger than 10000 may exceed
        # es result window size and doc_feeder maximum fetch size.
        # a number smaller than chunk_size is too small that the docs
        # can be sent to elasticsearch within one request, making it
        # inefficient, amplifying the scheduling overhead.

        x = IndexerCumulativeResult()
        for step in Step.order(steps):
            step = Step.dispatch(step)(self)
            self.logger.info(step)
            step.state.started()
            try:
                dx = await step.execute(job_manager, **kwargs)
                dx = IndexerStepResult(dx)
            except Exception as exc:
                _exc = str(exc)[:500]
                self.logger.exception(_exc)
                step.state.failed(_exc)
                raise exc
            else:
                merge(x.data, dx.data)
                self.logger.info(dx)
                self.logger.info(x)
                step.state.succeed(x.data)

        return x

    # ---------
    #   Steps
    # ---------
    async def pre_index(self, *args, mode, **kwargs):
        client = AsyncElasticsearch(**self.es_client_args)
        try:
            if mode in ("index", None):
                # index MUST NOT exist
                # ----------------------

                if await client.indices.exists(self.es_index_name):
                    msg = (
                        "Index '%s' already exists, (use mode='purge' to "
                        "auto-delete it or mode='resume' to add more documents)"
                    )
                    raise IndexerException(msg % self.es_index_name)

            elif mode in ("resume", "merge"):
                # index MUST exist
                # ------------------

                if not (await client.indices.exists(self.es_index_name)):
                    raise IndexerException("'%s' does not exist." % self.es_index_name)
                self.logger.info(("Exists", self.es_index_name))
                return  # skip index creation

            elif mode == "purge":
                # index MAY exist
                # -----------------

                response = await client.indices.delete(self.es_index_name, ignore_unavailable=True)
                self.logger.info(("Deleted", self.es_index_name, response))

            else:
                raise ValueError("Invalid mode: %s" % mode)

            response = await client.indices.create(
                self.es_index_name,
                body={
                    "settings": (await self.es_index_settings.finalize(client)),
                    "mappings": (await self.es_index_mappings.finalize(client)),
                },
            )
            self.logger.info(("Created", self.es_index_name, response))
            return {
                "__REPLACE__": True,
                "host": self.es_client_args.get("hosts"),  # for frontend display
                "environment": self.env_name,  # used in snapshot module.
            }

        finally:
            await client.close()

    async def do_index(self, job_manager, batch_size, ids, mode, **kwargs):
        client = DatabaseClient(**self.mongo_client_args)
        database = client[self.mongo_database_name]
        collection = database[self.mongo_collection_name]

        if ids:
            self.logger.info(
                ("Indexing from '%s' with specific list of _ids, " "create indexer job with batch_size=%d."),
                self.mongo_collection_name,
                batch_size,
            )
            # use user provided ids in batch
            id_provider = iter_n(ids, batch_size)
        else:
            self.logger.info(
                ("Fetch _ids from '%s', and create " "indexer job with batch_size=%d."),
                self.mongo_collection_name,
                batch_size,
            )
            # use ids from the target mongodb collection in batch
            id_provider = id_feeder(collection, batch_size, logger=self.logger)

        jobs = []  # asyncio.Future(s)
        error = None  # the first Exception

        total = len(ids) if ids else collection.count()
        schedule = Schedule(total, batch_size)

        def batch_finished(future):
            nonlocal error
            try:
                schedule.finished += future.result()
            except Exception as exc:
                self.logger.warning(exc)
                error = exc

        for batch_num, ids in zip(schedule, id_provider):
            await asyncio.sleep(0.0)

            # when one batch failed, and job scheduling has not completed,
            # stop scheduling and cancel all on-going jobs, to fail quickly.

            if error:
                for job in jobs:
                    if not job.done():
                        job.cancel()
                raise error

            self.logger.info(schedule)

            pinfo = self.pinfo.get_pinfo(schedule.suffix(self.mongo_collection_name))

            job = await job_manager.defer_to_process(
                pinfo,
                dispatch,
                self.mongo_client_args,
                self.mongo_database_name,
                self.mongo_collection_name,
                self.es_client_args,
                self.es_blkidx_args,
                self.es_index_name,
                ids,
                mode,
                batch_num,
            )
            job.add_done_callback(batch_finished)
            jobs.append(job)

        self.logger.info(schedule)
        await asyncio.gather(*jobs)

        schedule.completed()
        self.logger.notify(schedule)
        return {"count": total, "created_at": datetime.now().astimezone()}

    async def post_index(self, *args, **kwargs):
        ...


class ColdHotIndexer:
    """MongoDB to Elasticsearch 2-pass Indexer.
    (
        1st pass: <MongoDB Cold Collection>, # static data
        2nd pass: <MongoDB Hot Collection> # changing data
    ) =>
        <Elasticsearch Index>
    """

    # "ColdHotIndexer" is not a subclass of the "Indexer".
    # Step-level customization requires a subclass of "Indexer"
    # and assigning it to the "INDEXER" class attribute below.

    INDEXER = Indexer

    def __init__(self, build_doc, indexer_env, index_name):
        hot_build_doc = _BuildDoc(build_doc)
        cold_build_doc = hot_build_doc.extract_coldbuild()

        self.hot = self.INDEXER(hot_build_doc, indexer_env, index_name)
        self.cold = self.INDEXER(cold_build_doc, indexer_env, self.hot.es_index_name)

    async def index(
        self,
        job_manager,
        batch_size=10000,
        steps=("pre", "index", "post"),
        ids=None,
        mode=None,
        **kwargs,
    ):
        result = []

        cold_task = self.cold.index(
            job_manager,
            steps=set(Step.order(steps)) & {"pre", "index"},
            batch_size=batch_size,
            ids=ids,
            mode=mode,
        )
        result.append((await cold_task))

        hot_task = self.hot.index(
            job_manager,
            steps=set(Step.order(steps)) & {"index", "post"},
            batch_size=batch_size,
            ids=ids,
            mode="merge",
        )
        result.append((await hot_task))

        return result


class IndexManager(BaseManager):
    # An index config is considered a "source" for the manager
    # Each call returns a different instance from a factory call

    DEFAULT_INDEXER = Indexer

    def __init__(self, *args, **kwargs):
        """
        An example of config dict for this module.
        {
            "indexer_select": {
                None: "hub.dataindex.indexer.DrugIndexer", # default
                "build_config.cold_collection" : "mv.ColdHotVariantIndexer",
            },
            "env": {
                "prod": {
                    "host": "localhost:9200",
                    "indexer": {
                        "args": {
                            "timeout": 300,
                            "retry_on_timeout": True,
                            "max_retries": 10,
                        },
                        "bulk": {
                            "chunk_size": 50
                            "raise_on_exception": False
                        },
                        "concurrency": 3
                    },
                    "index": [
                        # for information only, only used in index_info
                        {"index": "mydrugs_current", "doc_type": "drug"},
                        {"index": "mygene_current", "doc_type": "gene"}
                    ],
                },
                "dev": { ... }
            }
        }
        """
        super().__init__(*args, **kwargs)
        self._srcbuild = get_src_build()
        self._config = {}

        self.logger, self.logfile = get_logger("indexmanager")

    # Object Lifecycle Calls
    # --------------------------
    # manager = IndexManager(job_manager)
    # manager.clean_stale_status() # in __init__
    # manager.configure(config)

    def clean_stale_status(self):
        IndexJobStateRegistrar.prune(get_src_build())

    def configure(self, conf):
        if not isinstance(conf, dict):
            raise TypeError(type(conf))
        self._config = conf

        # ES client argument defaults.
        # During heavy indexing, the following settings
        # significantly increase the one-pass success rate.
        esargs = {
            "timeout": 300,
            "retry_on_timeout": True,
            "max_retries": 10,
        }

        # register each indexing environment
        for name, env in conf["env"].items():
            indexer = env.setdefault("indexer", {})
            indexer.setdefault("args", dict(esargs))
            indexer["args"]["hosts"] = env.get("host")
            self.register[name] = deepcopy(indexer)
            self.register[name]["name"] = name

        self.logger.info(self.register)

    # Job Manager Hooks
    # ----------------------

    def get_predicates(self):
        def no_other_indexmanager_step_running(job_manager):
            """IndexManager deals with snapshot, publishing,
            none of them should run more than one at a time"""
            return len([j for j in job_manager.jobs.values() if j["category"] == INDEXMANAGER_CATEGORY]) == 0

        return [no_other_indexmanager_step_running]

    def get_pinfo(self):
        """
        Return dict containing information about the current process
        (used to report in the hub)
        """
        pinfo = {
            "category": INDEXMANAGER_CATEGORY,
            "source": "",
            "step": "",
            "description": "",
        }
        preds = self.get_predicates()
        if preds:
            pinfo["__predicates__"] = preds
        return pinfo

    # Hub Features
    # --------------

    def _select_indexer(self, build_name=None):
        """Find the indexer class required to index build_name."""

        rules = self._config.get("indexer_select")
        if not rules or not build_name:
            self.logger.debug(self.DEFAULT_INDEXER)
            return self.DEFAULT_INDEXER

        # the presence of a path in the build doc
        # can determine the indexer class to use.

        path = None
        doc = self._srcbuild.find_one({"_id": build_name})
        for path_in_doc, _ in traverse(doc or dict(), True):
            if path_in_doc in rules:
                if not path:
                    path = path_in_doc
                else:
                    _ERR = "Multiple indexers matched."
                    raise RuntimeError(_ERR)

        kls = get_class_from_classpath(rules[path])
        self.logger.info(kls)
        return kls

    def index(self, indexer_env, build_name, index_name=None, ids=None, **kwargs):
        """
        Trigger an index creation to index the collection build_name and create an
        index named index_name (or build_name if None). Optional list of IDs can be
        passed to index specific documents.
        """

        indexer_env_ = dict(self[indexer_env])  # describes a destination
        build_doc = self._srcbuild.find_one({"_id": build_name})  # describes a source

        if not build_doc:
            raise ValueError("Cannot find build %s." % build_name)
        if not build_doc.get("build_config"):
            raise ValueError("Cannot find build config for '%s'." % build_name)

        idx = self._select_indexer(build_name)
        idx = idx(build_doc, indexer_env_, index_name)
        job = idx.index(self.job_manager, ids=ids, **kwargs)
        job = asyncio.ensure_future(job)
        job.add_done_callback(self.logger.debug)

        return job

    def update_metadata(
        self,
        indexer_env,
        index_name,
        build_name=None,
        _meta=None,
    ):
        """
        Update _meta field of the index mappings, basing on
            1. the _meta value provided, including {}.
            2. the _meta value of the build_name in src_build.
            3. the _meta value of the build with the same name as the index.

        Examples:
            update_metadata("local", "mynews_201228_vsdevjd")
            update_metadata("local", "mynews_201228_vsdevjd", _meta={})
            update_metadata("local", "mynews_201228_vsdevjd", _meta={"author":"b"})
            update_metadata("local", "mynews_201228_current", "mynews_201228_vsdevjd")
        """

        async def _update_meta(_meta):
            env = self.register[indexer_env]
            async with AsyncElasticsearch(**env["args"]) as client:
                doc_type = None
                if int((await client.info())["version"]["number"].split(".")[0]) < 7:
                    mappings = client.indices.get_mapping(index_name)
                    mappings = mappings[index_name]["mappings"]
                    doc_type = next(iter(mappings.keys()))

                if _meta is None:
                    _id = build_name or index_name  # best guess
                    build = get_src_build().find_one({"_id": _id})
                    _meta = (build or {}).get("_meta")

                return await client.indices.put_mapping(
                    body=dict(_meta=_meta),
                    index=index_name,
                    doc_type=doc_type,
                )

        job = asyncio.ensure_future(_update_meta(_meta))
        job.add_done_callback(self.logger.debug)
        return job

    def index_info(self, remote=False):
        """Show index manager config with enhanced index information."""
        # http://localhost:7080/index_manager

        async def _enhance(conf):
            conf = copy.deepcopy(conf)

            for name, env in self.register.items():
                async with AsyncElasticsearch(**env["args"]) as client:
                    try:
                        indices = await client.indices.get("*")
                    except elasticsearch.exceptions.ConnectionError:
                        ...  # keep the hard-coded place-holders info
                    else:  # replace the index key with remote info
                        conf["env"][name]["index"] = [
                            {
                                "index": k,
                                "aliases": list(v["aliases"].keys()),
                                "doc_type": v["mappings"]["_meta"]["biothing_type"],
                            }
                            for k, v in indices.items()
                        ]
            return conf

        if remote:
            job = asyncio.ensure_future(_enhance(self._config))
            job.add_done_callback(self.logger.debug)
            return job

        return self._config

    def get_indexes_by_name(self, index_name=None, limit=10):
        """Accept an index_name and return a list of indexes get from all elasticsearch environments

        If index_name is blank, it will be return all indexes.
        limit can be used to specify how many indexes should be return.

        The list of indexes will be like this:
        [
            {
                "index_name": "...",
                "build_version": "...",
                "count": 1000,
                "creation_date": 1653468868933,
                "environment": {
                    "name": "env name",
                    "host": "localhost:9200",
                }
            },
        ]
        """

        if not index_name:
            index_name = "*"
        limit = int(limit)

        async def fetch(index_name, limit=None):
            indexes = []
            for env_name, env in self.register.items():
                async with AsyncElasticsearch(**env["args"]) as client:
                    try:
                        indices = await client.indices.get(index_name)
                    except Exception:
                        continue
                    for index_name, index_data in indices.items():
                        mapping_meta = index_data["mappings"]["_meta"]
                        indexes.append(
                            {
                                "index_name": index_name,
                                "build_version": mapping_meta["build_version"],
                                "count": mapping_meta["stats"]["total"],
                                "creation_date": index_data["settings"]["index"]["creation_date"],
                                "environment": {
                                    "name": env_name,
                                    "host": env["args"]["hosts"],
                                },
                            }
                        )

            indexes.sort(key=lambda index: index["creation_date"], reverse=True)

            if limit:
                indexes = indexes[:limit]
            return indexes

        job = asyncio.ensure_future(fetch(index_name, limit=limit))
        job.add_done_callback(self.logger.debug)
        return job

    def validate_mapping(self, mapping, env):
        indexer = self._select_indexer()  # default indexer
        indexer = indexer(dict(mapping=mapping), self[env], None)

        self.logger.debug(indexer.es_client_args)
        self.logger.debug(indexer.es_index_settings)
        self.logger.debug(indexer.es_index_mappings)

        async def _validate_mapping():
            client = AsyncElasticsearch(**indexer.es_client_args)
            index_name = ("hub_tmp_%s" % get_random_string()).lower()
            try:
                return await client.indices.create(
                    index_name,
                    body={
                        "settings": (await indexer.es_index_settings.finalize(client)),
                        "mappings": (await indexer.es_index_mappings.finalize(client)),
                    },
                )
            finally:
                await client.indices.delete(index_name, ignore_unavailable=True)
                await client.close()

        job = asyncio.ensure_future(_validate_mapping())
        job.add_done_callback(self.logger.info)
        return job

    def cleanup(self, env=None, keep=3, dryrun=True, **filters):
        """Delete old indices except for the most recent ones.

        Examples:
            >>> index_cleanup()
            >>> index_cleanup("production")
            >>> index_cleanup("local", build_config="demo")
            >>> index_cleanup("local", keep=0)
            >>> index_cleanup(_id="<elasticsearch_index>")
        """
        if not env and not dryrun:  # low specificity, unsafe.
            raise ValueError('Missing argument "env".')

        cleaner = Cleaner(get_src_build(), self, self.logger)
        cleanups = cleaner.find(env, keep, **filters)

        if dryrun:
            return "\n".join(
                (
                    "-" * 75,
                    cleaner.plain_text(cleanups),
                    "-" * 75,
                    "DRYRUN ONLY - APPLY THE ACTIONS WITH:",
                    "   > index_cleanup(..., dryrun=False)",
                )
            )

        job = asyncio.ensure_future(cleaner.clean(cleanups))
        job.add_done_callback(self.logger.info)
        return job


class DynamicIndexerFactory:
    """
    In the context of autohub/standalone instances, create indexer
    with parameters taken from versions.json URL.
    A list of  URLs is provided so the factory knows how to create these
    indexers for each URLs. There's no way to "guess" an ES host from a URL,
    so this parameter must be specified as well, common to all URLs
    "suffix" param is added at the end of index names.
    """

    def __init__(self, urls, es_host, suffix="_current"):
        self.urls = urls
        self.es_host = es_host
        self.bynames = {}
        for url in urls:
            if isinstance(url, dict):
                name = url["name"]
                # actual_url = url["url"]
            else:
                name = os.path.basename(os.path.dirname(url))
                # actual_url = url
            self.bynames[name] = {
                "es_host": self.es_host,
                "index": name + suffix,
            }

    def create(self, name):
        conf = self.bynames[name]
        pidxr = partial(
            ESIndexer,
            index=conf["index"],
            doc_type=None,
            es_host=conf["es_host"],
        )
        conf = {"es_host": conf["es_host"], "index": conf["index"]}
        return pidxr, conf



from functools import wraps
from types import CoroutineType

import flask
from tornado.template import Loader

from biothings.web import templates
from biothings.web.options import OptionError
from biothings.web.query.pipeline import QueryPipelineException, QueryPipelineInterrupt

routes = []


def route(pattern, methods=("GET", "POST")):
    def A(f):
        async def B(*args, **kwargs):
            biothings = flask.current_app.biothings
            optionsets = biothings.optionsets
            optionset = optionsets.get(f.__name__)
            if optionset:
                try:
                    _args = optionset.parse(
                        flask.request.method,
                        (
                            (tuple(kwargs.values()), {}),
                            flask.request.args,
                            flask.request.form,
                            flask.request.get_json(),
                        ),
                    )
                except OptionError as err:
                    return err.info, 400
            else:
                _args = {}
            result = f(biothings, _args)
            if isinstance(result, CoroutineType):
                return await result
            return result

        B.pattern = pattern
        B.methods = methods
        B.name = f.__name__
        routes.append(B)
        return B

    return A


@route("/")
def homepage(biothings, args):
    loader = Loader(templates.__path__[0])
    template = loader.load("home.html")
    return template.generate(
        alert="Front Page Not Configured.",
        title="Biothings API",
        contents=biothings.handlers.keys(),
        support=biothings.metadata.types,
        url="http://biothings.io/",
    )


def handle_es_conn(f):
    @wraps(f)
    async def _(biothings, *args, **kwargs):
        client = biothings.elasticsearch.async_client
        # because of the flask execution model
        # each time the async function is executed
        # it is executed on a different event loop
        # reset the connections to use the active loop
        del client.transport.connection_pool
        await client.transport._async_init()
        try:
            response = await f(biothings, *args, **kwargs)
        except QueryPipelineInterrupt as itr:
            return itr.details
        except QueryPipelineException as exc:
            kwargs = exc.details if isinstance(exc.details, dict) else {}
            kwargs["success"] = False
            kwargs["status"] = exc.code
            kwargs["reason"] = exc.summary
            return kwargs, exc.code
        finally:
            await client.close()
        return response

    return _


@route("/{ver}/query")
@handle_es_conn
async def query(biothings, args):
    return await biothings.pipeline.search(**args)


@route(["/{ver}/{typ}/", "/{ver}/{typ}/<id>"])
@handle_es_conn
async def annotation(biothings, args):
    # could be a list, in which case we need jsonify.
    return flask.jsonify(await biothings.pipeline.fetch(**args))


@route("/{ver}/metadata")
@handle_es_conn
async def metadata(biothings, args):
    await biothings.metadata.refresh(None)
    return biothings.metadata.get_metadata(None)


@route("/{ver}/metadata/fields")
@handle_es_conn
async def fields(biothings, args):
    await biothings.metadata.refresh(None)
    mappings = biothings.metadata.get_mappings(None)
    return biothings.pipeline.formatter.transform_mapping(mappings)


@route("/status")
@handle_es_conn
async def status(biothings, args):
    return await biothings.health.async_check()



# Proof of concept
# Not production ready.

# These routes represent a native implementaton
# of the biothings apis with limited functionality

# A comprehensive translation of all biothings
# default parameters may require building the
# handler functions with FunctionType and AST
# https://docs.python.org/3.8/library/ast.html

from types import SimpleNamespace

from fastapi import HTTPException

from biothings.web.query.pipeline import QueryPipelineException, QueryPipelineInterrupt

routes = []
biothings = SimpleNamespace()


def route(*args, **kwargs):
    def _(f):
        f.name = f.__name__
        f.args = args
        f.kwargs = kwargs
        routes.append(f)
        return f

    return _


async def _capture_exc(coro):
    try:
        return await coro
    except QueryPipelineInterrupt as itr:
        return itr.details
    except QueryPipelineException as exc:
        kwargs = exc.details if isinstance(exc.details, dict) else {}
        kwargs["status"] = exc.code
        kwargs["reason"] = exc.summary
        raise HTTPException(exc.code, kwargs)


@route("/")
async def root():
    return {"message": "Hello World"}


@route("/v1/query")
async def query(q="__all__"):
    return await _capture_exc(biothings.db.pipeline.search(q))


@route("/v1/doc/{id}")
async def annotation(id):
    return await _capture_exc(biothings.db.pipeline.fetch(id))


@route("/v1/metadata")
async def metadata():
    await biothings.metadata.refresh(None)
    return biothings.metadata.get_metadata(None)


@route("/v1/metadata/fields")
async def fields():
    await biothings.metadata.refresh(None)
    mappings = biothings.metadata.get_mappings(None)
    return biothings.db.pipeline.formatter.transform_mapping(mappings)
